'''
Created on 29.03.2017

@author: abaktheer

Microplane Fatigue model 2D

(compressive plasticity (CP) + tensile damage (TD) 
+ cumulative damage sliding (CSD))

Using Jirasek homogenization approach [1999]
'''

from traits.api import Constant, \
    Float, Dict, Property, cached_property

from ibvpy.tmodel.mats2D.mats2D_eval import MATS2DEval
import numpy as np
import traits.api as tr


class MATS2DMplCSDEEQ(MATS2DEval):

    #---------------------------------------
    # Tangential constitutive law parameters
    #---------------------------------------
    gamma_T = Float(5000.,
                    label="Gamma",
                    desc=" Tangential Kinematic hardening modulus",
                    enter_set=True,
                    auto_set=False)

    K_T = Float(10.0,
                label="K",
                desc="Tangential Isotropic harening",
                enter_set=True,
                auto_set=False)

    S_T = Float(0.00001,
                label="S",
                desc="Damage strength",
                enter_set=True,
                auto_set=False)

    r_T = Float(1.2,
                label="r",
                desc="Damage cumulation parameter",
                enter_set=True,
                auto_set=False)

    c_T = Float(1.2,
                label="c",
                desc="Damage cumulation parameter",
                enter_set=True,
                auto_set=False)

    tau_pi_bar = Float(2.0,
                       label="Tau_bar",
                       desc="Reversibility limit",
                       enter_set=True,
                       auto_set=False)

    a = Float(0.0,
              label="a",
              desc="Lateral pressure coefficient",
              enter_set=True,
              auto_set=False)

    #-------------------------------------------
    # Normal_Tension constitutive law parameters (without cumulative normal strain)
    #-------------------------------------------
    Ad = Float(10000.0,
               label="a",
               desc="brittleness coefficient",
               enter_set=True,
               auto_set=False)

    eps_0 = Float(0.0002,
                  label="a",
                  desc="threshold strain",
                  enter_set=True,
                  auto_set=False)


    #-----------------------------------------------
    # Normal_Compression constitutive law parameters
    #-----------------------------------------------
    K_N = Float(10000.,
                    label="K_N",
                    desc=" Normal isotropic harening",
                    enter_set=True,
                    auto_set=False)

    gamma_N = Float(15000.,
                        label="gamma_N",
                        desc="Normal kinematic hardening",
                        enter_set=True,
                        auto_set=False)

    sigma_0 = Float(20.,
                        label="sigma_0",
                        desc="Yielding stress",
                        enter_set=True,
                        auto_set=False)

    state_var_shapes = Property(Dict(), depends_on='n_mp')
    '''Dictionary of state variable entries with their array shapes.
    '''
    #-------------------------------------------------------------------------
    # Cached elasticity tensors
    #-------------------------------------------------------------------------

    E = tr.Float(34e+3,
                 label="E",
                 desc="Young's Modulus",
                 auto_set=False,
                 input=True)

    nu = tr.Float(0.2,
                  label='nu',
                  desc="Poison ratio",
                  auto_set=False,
                  input=True)

    def _get_lame_params(self):
        la = self.E * self.nu / ((1. + self.nu) * (1. - 2. * self.nu))
        # second Lame parameter (shear modulus)
        mu = self.E / (2. + 2. * self.nu)
        return la, mu

    D_abef = tr.Property(tr.Array, depends_on='+input')

    @tr.cached_property
    def _get_D_abef(self):
        la = self._get_lame_params()[0]
        mu = self._get_lame_params()[1]
        delta = np.identity(2)
        D_abef = (np.einsum(',ij,kl->ijkl', la, delta, delta) +
                  np.einsum(',ik,jl->ijkl', mu, delta, delta) +
                  np.einsum(',il,jk->ijkl', mu, delta, delta))

        return D_abef

    @cached_property
    def _get_state_var_shapes(self):
        return {'w_N_Emn': (self.n_mp,),
                'z_N_Emn': (self.n_mp,),
                'alpha_N_Emn': (self.n_mp,),
                'r_N_Emn': (self.n_mp,),
                'eps_N_p_Emn': (self.n_mp,),
                'sigma_N_Emn': (self.n_mp,),
                'w_T_Emn': (self.n_mp,),
                'z_T_Emn': (self.n_mp,),
                'alpha_T_Emna': (self.n_mp, 2),
                'eps_T_pi_Emna': (self.n_mp, 2),
                }

    #--------------------------------------------------------------
    # microplane constitutive law (normal behavior CP + TD)
    # (without cumulative normal strain for fatigue under tension)
    #--------------------------------------------------------------
    def get_normal_law(self, eps_N_Emn, w_N_Emn, z_N_Emn,
                       alpha_N_Emn, r_N_Emn, eps_N_p_Emn):

        E_N = self.E / (1.0 - 2.0 * self.nu)

        pos = eps_N_Emn > 1e-6
        H = 1.0 * pos

        sigma_n_trial = (1.0 - H * w_N_Emn) * E_N * (eps_N_Emn - eps_N_p_Emn)
        Z = self.K_N * r_N_Emn
        X = self.gamma_N* alpha_N_Emn
        h = self.sigma_0+ Z
        pos_iso = h > 1e-6
        f_trial = abs(sigma_n_trial - X) - h * pos_iso

        thres_1 = f_trial > 1e-6

        delta_lamda = f_trial / \
            (E_N + abs(self.K_N) + self.gamma_N) * thres_1
        eps_N_p_Emn = eps_N_p_Emn + delta_lamda * np.sign(sigma_n_trial - X)
        r_N_Emn = r_N_Emn + delta_lamda
        alpha_N_Emn = alpha_N_Emn + delta_lamda * np.sign(sigma_n_trial - X)

        def Z_N(z_N_Emn): return 1.0 / self.Ad * (-z_N_Emn) / (1.0 + z_N_Emn)

        Y_N = 0.5 * H * E_N * eps_N_Emn ** 2.0
        Y_0 = 0.5 * E_N * self.eps_0 ** 2.0
        f = Y_N - (Y_0 + Z_N(z_N_Emn))

        thres_2 = f > 1e-6

        def f_w(Y): return 1.0 - 1.0 / (1.0 + self.Ad * (Y - Y_0))

        w_N_Emn = f_w(Y_N) * thres_2
        z_N_Emn = -w_N_Emn * thres_2


        sigma_N_Emn = (1.0 - H * w_N_Emn) * E_N * (eps_N_Emn - eps_N_p_Emn)

        return w_N_Emn, z_N_Emn, alpha_N_Emn, r_N_Emn, eps_N_p_Emn, sigma_N_Emn


    #-------------------------------------------------------------------------
    # microplane constitutive law (Tangential CSD)-(Pressure sensitive cumulative damage)
    #-------------------------------------------------------------------------
    def get_tangential_law(self, eps_T_Emna, w_T_Emn, z_T_Emn,
                           alpha_T_Emna, eps_T_pi_Emna, sigma_N_Emn):

        E_T = self.E / (1.0 + self.nu)

        sig_pi_trial = E_T * (eps_T_Emna - eps_T_pi_Emna)
        Z = self.K_T * z_T_Emn
        X = self.gamma_T * alpha_T_Emna
        norm_1 = np.sqrt(
            np.einsum(
                '...na,...na->...n',
                (sig_pi_trial - X), (sig_pi_trial - X))
        )

        f = norm_1 - self.tau_pi_bar - \
            Z + self.a * sigma_N_Emn / 3.0

        plas_1 = f > 1e-6
        elas_1 = f < 1e-6

        delta_lamda = f / \
            (E_T / (1.0 - w_T_Emn) + self.gamma_T + self.K_T) * plas_1

        norm_2 = 1.0 * elas_1 + np.sqrt(
            np.einsum(
                '...na,...na->...n',
                (sig_pi_trial - X), (sig_pi_trial - X))) * plas_1

        eps_T_pi_Emna[..., 0] = eps_T_pi_Emna[..., 0] + plas_1 * delta_lamda * \
            ((sig_pi_trial[..., 0] - X[..., 0]) /
             (1.0 - w_T_Emn)) / norm_2
        eps_T_pi_Emna[..., 1] = eps_T_pi_Emna[..., 1] + plas_1 * delta_lamda * \
            ((sig_pi_trial[..., 1] - X[..., 1]) /
             (1.0 - w_T_Emn)) / norm_2

        Y = 0.5 * E_T * \
            np.einsum(
                '...na,...na->...n',
                (eps_T_Emna - eps_T_pi_Emna),
                (eps_T_Emna - eps_T_pi_Emna)
            )

        w_T_Emn += ((1 - w_T_Emn) ** self.c_T) * \
            (delta_lamda * (Y / self.S_T) ** self.r_T) * \
            (self.tau_pi_bar / (self.tau_pi_bar - self.a * sigma_N_Emn / 3.0))

        alpha_T_Emna[..., 0] = alpha_T_Emna[..., 0] + plas_1 * delta_lamda * \
            (sig_pi_trial[..., 0] - X[..., 0]) / norm_2
        alpha_T_Emna[..., 1] = alpha_T_Emna[..., 1] + plas_1 * delta_lamda * \
            (sig_pi_trial[..., 1] - X[..., 1]) / norm_2

        z_T_Emn = z_T_Emn + delta_lamda

        return w_T_Emn, z_T_Emn, alpha_T_Emna, eps_T_pi_Emna

#     #-------------------------------------------------------------------------
#     # MICROPLANE-Kinematic constraints
#     #-------------------------------------------------------------------------
#     def _get_e_Emna(self, eps_Emab):
#         # Projection of apparent strain onto the individual microplanes
#         e_ni = np.einsum('nb,Emba->Emna', self._MPN, eps_Emab)
#         return e_ni
# 
#     def _get_e_N_Emn(self, e_Emna):
#         # get the normal strain array for each microplane
# 
#         e_N_Emn = np.einsum('nij,...ij->...n', e_Emna, self._MPN)
#         return e_N_Emn
# 
#     def _get_e_T_Emna(self, e_Emna):
#         # get the tangential strain vector array for each microplane
#         MPTT_ijr = self._get__MPTT()
#         return np.einsum('nija,...ij->...na', MPTT_ijr, e_Emna)

    #-------------------------------------------------
    # Alternative methods for the kinematic constraint
    #-------------------------------------------------
    
    # get the operator of the microplane normals
    _MPNN = Property(depends_on='n_mp')
    
    @cached_property
    def _get__MPNN(self):
        MPNN_nij = np.einsum('ni,nj->nij', self._MPN, self._MPN)
        return MPNN_nij

    # get the third order tangential tensor (operator) for each microplane
    _MPTT = Property(depends_on='n_mp')
    
    @cached_property
    def _get__MPTT(self):
        delta = np.identity(2)
        MPTT_nijr = 0.5 * (
            np.einsum('ni,jr -> nijr', self._MPN, delta) +
            np.einsum('nj,ir -> njir', self._MPN, delta) - 2 *
            np.einsum('ni,nj,nr -> nijr', self._MPN, self._MPN, self._MPN)
        )
        return MPTT_nijr

    def _get_e_N_Emn_2(self, eps_Emab):
        # get the normal strain array for each microplane
        return np.einsum('nij,...ij->...n', self._MPNN, eps_Emab)

    def _get_e_T_Emnar_2(self, eps_Emab):
        # get the tangential strain vector array for each microplane
        MPTT_ijr = self._get__MPTT()
        return np.einsum('nija,...ij->...na', MPTT_ijr, eps_Emab)

    #--------------------------------------------------------
    # return the state variables (Damage , inelastic strains)
    #--------------------------------------------------------
    def _get_state_variables(self, eps_Emab, tn1,
                             omegaN, z_N_Emn,
                             alpha_N_Emn, r_N_Emn, eps_N_p_Emn, sigma_N_Emn,
                             w_T_Emn, z_T_Emn, alpha_T_Emna, eps_T_pi_Emna):
 
        e_N_arr = self._get_e_N_Emn_2(eps_Emab)
        e_T_vct_arr = self._get_e_T_Emnar_2(eps_Emab)
 
        omegaN, z_N_Emn, alpha_N_Emn, r_N_Emn, eps_N_p_Emn, sigma_N_Emn = self.get_normal_law(e_N_arr,  omegaN, z_N_Emn,
                                                                                              alpha_N_Emn, r_N_Emn, eps_N_p_Emn)
 
        w_T_Emn, z_T_Emn, alpha_T_Emna, eps_T_pi_Emna = self.get_tangential_law(e_T_vct_arr, w_T_Emn, z_T_Emn,
                                                                                alpha_T_Emna, eps_T_pi_Emna, sigma_N_Emn)
 
        return omegaN, z_N_Emn, alpha_N_Emn, r_N_Emn, eps_N_p_Emn, sigma_N_Emn, w_T_Emn, z_T_Emn, alpha_T_Emna, eps_T_pi_Emna

    #-----------------------------------------------------------------
    # Returns a list of the plastic normal strain  for all microplanes.
    #-----------------------------------------------------------------
    def _get_eps_N_p_Emn(self, eps_Emab, w_N_Emn, z_N_Emn, alpha_N_Emn, r_N_Emn, eps_N_p_Emn, sigma_N_Emn):
        
        eps_N_Emn = self._get_e_N_Emn_2(eps_Emab)
        
        eps_N_p_Emn = self.get_normal_law(
            eps_N_Emn, w_N_Emn, z_N_Emn, alpha_N_Emn, r_N_Emn, eps_N_p_Emn, sigma_N_Emn)[4]
            
        return eps_N_p_Emn

    #----------------------------------------------------------------
    # Returns a list of the sliding strain vector for all microplanes.
    #----------------------------------------------------------------
    def _get_eps_T_pi_arr(self, eps_Emab, w_T_Emn, z_T_Emn, alpha_T_Emna, eps_T_pi_Emna, sigma_N_Emn):

        eps_T_Emna = self._get_e_T_Emnar_2(eps_Emab)
        
        eps_N_T_pi_Emna = self.get_tangential_law(
            eps_T_Emna, w_T_Emn, z_T_Emn, alpha_T_Emna, eps_T_pi_Emna, sigma_N_Emn)[3]
           
        return eps_N_T_pi_Emna

    #-------------------------------------------------------------
    # Returns a list of the integrity factors for all microplanes.
    #-------------------------------------------------------------
    def _get_phi_Emn(self, eps_Emab, w_N_Emn, z_N_Emn,
                     alpha_N_Emn, r_N_Emn, eps_N_p_Emn,
                     w_T_Emn, z_T_Emn, alpha_T_Emna, eps_T_pi_Emna, sigma_N_Emn):

        eps_N_Emn = self._get_e_N_Emn_2(eps_Emab)
        eps_T_Emna = self._get_e_T_Emnar_2(eps_Emab)

        w_N_Emn = self.get_normal_law(
            eps_N_Emn, w_N_Emn, z_N_Emn, alpha_N_Emn, r_N_Emn, eps_N_p_Emn)[0]
            
        w_T_Emn = self.get_tangential_law(
            eps_T_Emna, w_T_Emn, z_T_Emn, alpha_T_Emna, eps_T_pi_Emna, sigma_N_Emn)[0]

        w_Emn = np.zeros_like(w_N_Emn)

        #w_Emn = np.maximum(w_N_Emn, w_T_Emn)

        w_Emn = w_T_Emn
        #print('w_N_Emn', w_N_Emn)
        #print('w_T_Emn', w_T_Emn)
        #print('w_Emn', w_Emn)

        phi_Emn = np.sqrt(1.0 - w_Emn)
        #print('phi_Emn', phi_Emn)

        return phi_Emn

    #----------------------------------------------
    # Returns the 2nd order damage tensor 'phi_mtx'
    #----------------------------------------------
    def _get_phi_Emab(self, eps_Emab, w_N_Emn, z_N_Emn,
                      alpha_N_Emn, r_N_Emn, eps_N_p_Emn,
                      w_T_Emn, z_T_Emn, alpha_T_Emna, eps_T_pi_Emna, sigma_N_Emn):

        # scalar integrity factor for each microplane
        phi_Emn = self._get_phi_Emn(eps_Emab, w_N_Emn, z_N_Emn,
                                    alpha_N_Emn, r_N_Emn, eps_N_p_Emn,
                                    w_T_Emn, z_T_Emn, alpha_T_Emna, eps_T_pi_Emna, sigma_N_Emn)

        # integration terms for each microplanes
        phi_Emab = np.einsum('...n,n,nab->...ab', phi_Emn,
                             self._MPW, self._MPNN)

        return phi_Emab

    #----------------------------------------------------------------------
    # Returns the 4th order damage tensor 'beta4' using sum-type symmetrization
    # (cf. [Jir99], Eq.(21))
    #----------------------------------------------------------------------
    def _get_beta_Emabcd(self, eps_Emab, w_N_Emn, z_N_Emn,
                         alpha_N_Emn, r_N_Emn, eps_N_p_Emn,
                         w_T_Emn, z_T_Emn, alpha_T_Emna, eps_T_pi_Emna, sigma_N_Emn):

        delta = np.identity(2)

        phi_Emab = self._get_phi_Emab(eps_Emab, w_N_Emn, z_N_Emn,
                                      alpha_N_Emn, r_N_Emn, eps_N_p_Emn,
                                      w_T_Emn, z_T_Emn, alpha_T_Emna, eps_T_pi_Emna, sigma_N_Emn)

        # use numpy functionality (einsum) to evaluate [Jir99], Eq.(21)
        beta_Emabcd = 0.25 * (
            np.einsum('...ik,jl->...ijkl', phi_Emab, delta) +
            np.einsum('...il,jk->...ijkl', phi_Emab, delta) +
            np.einsum('...jk,il->...ijkl', phi_Emab, delta) +
            np.einsum('...jl,ik->...ijkl', phi_Emab, delta)
        )

        return beta_Emabcd

    #---------------------------------------------------------------------
    # Extra homogenization of damage tensor in case of two damage parameters
    # Returns the 4th order damage tensor 'beta4' using (ref. [Baz99], Eq.(63))
    #---------------------------------------------------------------------

    def _get_beta_Emabcd_2(self, eps_Emab, w_N_Emn, z_N_Emn,
            alpha_N_Emn, r_N_Emn, eps_N_p_Emn, w_T_Emn, z_T_Emn,
             alpha_T_Emna, eps_T_pi_Emna, sigma_N_Emn):

        # Returns the 4th order damage tensor 'beta4' using
        #(cf. [Baz99], Eq.(63))
        
        eps_N_Emn = self._get_e_N_Emn_2(eps_Emab)
        eps_T_Emna = self._get_e_T_Emnar_2(eps_Emab)
        
        w_N_Emn = self.get_normal_law(
            eps_N_Emn, w_N_Emn, z_N_Emn, alpha_N_Emn, r_N_Emn, eps_N_p_Emn)[0]
            
        w_T_Emn = self.get_tangential_law(
            eps_T_Emna, w_T_Emn, z_T_Emn, alpha_T_Emna, eps_T_pi_Emna, sigma_N_Emn)[0]

        delta = np.identity(2)
        beta_N = np.sqrt(1. - w_N_Emn)
        beta_T = np.sqrt(1. - w_T_Emn)

        #beta_N = 1. - w_N_Emn
        #beta_T = 1. - w_T_Emn

        #print(' w_N_Emn ',  w_N_Emn)
        #print(' w_N_Emn ', w_T_Emn)
        #print('beta_N ', beta_N)
        #print('beta_T ', beta_T)

        beta_ijkl = np.einsum('n, ...n,ni, nj, nk, nl -> ...ijkl', self._MPW, beta_N, self._MPN, self._MPN, self._MPN, self._MPN) + \
            0.25 * (np.einsum('n, ...n,ni, nk, jl -> ...ijkl', self._MPW, beta_T, self._MPN, self._MPN, delta) +
                    np.einsum('n, ...n,ni, nl, jk -> ...ijkl', self._MPW, beta_T, self._MPN, self._MPN, delta) +
                    np.einsum('n, ...n,nj, nk, il -> ...ijkl', self._MPW, beta_T, self._MPN, self._MPN, delta) +
                    np.einsum('n, ...n,nj, nl, ik -> ...ijkl', self._MPW, beta_T, self._MPN, self._MPN, delta) -
                    4.0 * np.einsum('n, ...n, ni, nj, nk, nl -> ...ijkl', self._MPW, beta_T, self._MPN, self._MPN, self._MPN, self._MPN))

        return beta_ijkl
    #-----------------------------------------------------------
    # Integration of the (inelastic) strains for each microplane
    #-----------------------------------------------------------

    def _get_eps_p_Emab(self, eps_Emab, w_N_Emn, z_N_Emn,
                        alpha_N_Emn, r_N_Emn, eps_N_p_Emn,
                        w_T_Emn, z_T_Emn, alpha_T_Emna, eps_T_pi_Emna, sigma_N_Emn):

        eps_N_Emn = self._get_e_N_Emn_2(eps_Emab)
        eps_T_Emna = self._get_e_T_Emnar_2(eps_Emab)

        # plastic normal strains
        eps_N_p_Emn = self.get_normal_law(
            eps_N_Emn, w_N_Emn, z_N_Emn, alpha_N_Emn, r_N_Emn, eps_N_p_Emn)[4]

        # sliding tangential strains
        eps_T_pi_Emna = self.get_tangential_law(
            eps_T_Emna, w_T_Emn, z_T_Emn, alpha_T_Emna, eps_T_pi_Emna, sigma_N_Emn)[3]
            
        print('eps_N_T_pi_Emna ' ,eps_T_pi_Emna ) 
         
        delta = np.identity(2)

        # 2-nd order plastic (inelastic) tensor
        eps_p_Emab = (
            np.einsum('n,...n,na,nb->...ab',
                      self._MPW, eps_N_p_Emn, self._MPN, self._MPN) +
            0.5 * (
                np.einsum('n,...nf,na,fb->...ab',
                          self._MPW, eps_T_pi_Emna, self._MPN, delta) +
                np.einsum('n,...nf,nb,fa->...ab', self._MPW,
                          eps_T_pi_Emna, self._MPN, delta)
            )
        )

        return eps_p_Emab

    #-------------------------------------------------------------------------
    # Evaluation - get the corrector and predictor
    #-------------------------------------------------------------------------

    def get_corr_pred(self, eps_Emab, t_n1, w_N_Emn, z_N_Emn,
                      alpha_N_Emn, r_N_Emn, eps_N_p_Emn, sigma_N_Emn,
                      w_T_Emn, z_T_Emn, alpha_T_Emna, eps_T_pi_Emna):

        # Corrector predictor computation.

        #------------------------------------------------------------------
        # Damage tensor (4th order) using product- or sum-type symmetrization:
        #------------------------------------------------------------------
        beta_Emabcd = self._get_beta_Emabcd(
            eps_Emab, w_N_Emn, z_N_Emn,
            alpha_N_Emn, r_N_Emn, eps_N_p_Emn,
            w_T_Emn, z_T_Emn, alpha_T_Emna, eps_T_pi_Emna, sigma_N_Emn
        )

        #------------------------------------------------------------------
        # Damaged stiffness tensor calculated based on the damage tensor beta4:
        #------------------------------------------------------------------

        D_Emabcd = np.einsum(
            '...ijab, abef, ...cdef->...ijcd', beta_Emabcd, self.D_abef, beta_Emabcd)
        #----------------------------------------------------------------------
        # Return stresses (corrector) and damaged secant stiffness matrix (predictor)
        #----------------------------------------------------------------------
        # plastic strain tensor
        eps_p_Emab = self._get_eps_p_Emab(
            eps_Emab, w_N_Emn, z_N_Emn,
            alpha_N_Emn, r_N_Emn, eps_N_p_Emn,
            w_T_Emn, z_T_Emn, alpha_T_Emna, eps_T_pi_Emna, sigma_N_Emn)

        # elastic strain tensor
        eps_e_Emab = eps_Emab - eps_p_Emab

        # calculation of the stress tensor
        sig_Emab = np.einsum('...abcd,...cd->...ab', D_Emabcd, eps_e_Emab)

        return D_Emabcd, sig_Emab


# class MATS2DMplCSDEEQ(MATSXDMplCDSEEQ, MATS2DEval):

    # implements(IMATSEval)

    #-----------------------------------------------
    # number of microplanes 
    #-----------------------------------------------
    n_mp = Constant(360)

    #-----------------------------------------------
    # get the normal vectors of the microplanes
    #-----------------------------------------------
    _MPN = Property(depends_on='n_mp')

    @cached_property
    def _get__MPN(self):
        # microplane normals:
        alpha_list = np.linspace(0, 2 * np.pi, self.n_mp)

        MPN = np.array([[np.cos(alpha), np.sin(alpha)]
                        for alpha in alpha_list])

        return MPN

    #-------------------------------------
    # get the weights of the microplanes
    #-------------------------------------
    _MPW = Property(depends_on='n_mp')

    @cached_property
    def _get__MPW(self):
        MPW = np.ones(self.n_mp) / self.n_mp * 2

        return MPW



'''
Created on Dec 17, 2018

@author: rch
'''
from traits.api import Interface


class ITStep(Interface):

    def make_iter(self):
        '''Perform one iteration step.
        '''

    def make_incr(self):
        '''Update the control, primary and state variables..
        '''



'''
Created on Feb 14, 2019

@author: rch
'''

from ibvpy.tmodel import MATSEval
from ibvpy.tmodel.mats_damage_fn import \
    IDamageFn, LiDamageFn, ExpSlopeDamageFn, AbaqusDamageFn, \
    MultilinearDamageFn, \
    FRPDamageFn
from ibvpy.sim.i_tmodel import ITModel
from traits.api import on_trait_change
import numpy as np
import traits.api as tr
import bmcs_utils.api as bu

@tr.provides(ITModel)
class MATS1D5D(MATSEval):

    node_name = "damage bond model"

    E_T = tr.Float(100.0, tooltip='Shear stiffness of the interface [MPa]',
                   MAT=True, unit='MPa', symbol='E_\mathrm{s}',
                   desc='Shear-modulus of the interface',
                   auto_set=True, enter_set=True)

    E_N = tr.Float(100.0, tooltip='Normal stiffness of the interface [MPa]',
                   MAT=True, unit='MPa', symbol='E_\mathrm{n}',
                   desc='Normal stiffness of the interface',
                   auto_set=False, enter_set=True)

    state_var_shapes = dict(omega=(),
                            kappa=())

    omega_fn_type = tr.Trait('exp-slope',
                             dict(li=LiDamageFn,
                                  exp_slope=ExpSlopeDamageFn,
                                  abaqus=AbaqusDamageFn,
                                  FRP=FRPDamageFn,
                                  multilinear=MultilinearDamageFn
                                  ),
                             MAT=True,
                             )

    @on_trait_change('omega_fn_type')
    def _reset_omega_fn(self):
        self.omega_fn = self.omega_fn_type_()

    omega_fn = tr.Instance(IDamageFn, report=True)

    def _omega_fn_default(self):
        return self.omega_fn_type_()

    def omega(self, k):
        return self.omega_fn(k)

    def omega_derivative(self, k):
        return self.omega_fn.diff(k)

    def init(self, omega, kappa):
        r'''
        Initialize the state variables.
        '''
        omega[...] = 0
        kappa[...] = 0

    algorithmic = tr.Bool(True)

    def get_corr_pred(self, u_r, t_n, omega, kappa):

        s = u_r[..., 0]
        w = u_r[..., 1]
        # For normal
        H_w_N = np.array(w <= 0.0, dtype=np.float64)
        E_alg_N = H_w_N * self.E_N
        sig_N = E_alg_N * w
        kappa[...] = np.max(np.array([kappa, np.fabs(s)]), axis=0)
        omega[...] = self.omega(kappa)
        tau = (1 - omega) * self.E_T * s
        domega_ds = self.omega_derivative(kappa)

        if self.algorithmic:
            E_alg_T = ((1 - omega) - domega_ds * s) * self.E_T
        else:
            E_alg_T = (1 - omega) * self.E_T

        sig = np.zeros_like(u_r)
        sig[..., 0] = tau
        sig[..., 1] = sig_N
        E_TN = np.einsum('ab...->...ab',
            np.array(
             [
                 [E_alg_T, np.zeros_like(E_alg_T)],
                 [np.zeros_like(E_alg_N), E_alg_N]
             ])
            )
        return sig, E_TN

    ipw_view = bu.View(
        bu.Item('E_T'),
        bu.Item('E_N')
    )




from traits.api import Array, Int, List, Interface, Type

#-------------------------------------------------------------------
# IFETSEval - interface for fe-numerical quadrature
#-------------------------------------------------------------------

class IFElem( Interface ):
    '''Finite Element interface for spatial resolution and integration.
    '''
    dof_r = List
    
    geo_r = List
    
    n_nodal_dofs = Int
    

class IFETSEval( IFElem ):
    '''Interface for time steppers.
    Unified interface for spatial and temporal resolution and integration.
    '''
    #        
    #    The class of the domain time-stepper must be harmonized with the 
    #    integration scheme over a finite elements. Essentially, two types are 
    #    distinguished:
    #    - regular integration implemented in DOTSEval
    #    - irregular integration (mainteinance of integration points for
    #      each element within the domain is required) implemented in XDOTSEval
    #
    dots_class = Type

    def adjust_spatial_context_for_point( self, sctx ):
        '''
        Method gets called prior to the evaluation at the material point level.
        
        The method can be used for dimensionally reduced elements that 
        are using higher-dimensional material models.
        '''        
        pass
    
    def get_corr_pred(self, sctx, u, du, tn, tn1, u_avg, 
                      B_mtx_grid, J_det_grid,
                      ip_coords, ip_weights ):   
        '''Return the corrector and predictor for the next step.
        ''' 
        pass



from ibvpy.mathkit.linalg.sys_mtx_assembly import SysMtxArray
from ibvpy.mathkit.tensor import EPS, DELTA
from ibvpy.sim.i_xmodel import IXModel
from traits.api import \
    provides, Property
from ibvpy.view.ui import BMCSLeafNode

import numpy as np


DD = np.hstack([DELTA, np.zeros_like(DELTA)])
EEPS = np.hstack([np.zeros_like(EPS), EPS])
GAMMA = np.einsum(
    'ik,jk->kij', DD, DD
) + np.einsum(
    'ikj->kij', np.fabs(EEPS)
)
GAMMA_inv = np.einsum(
    'ik,jk->kij', DD, DD
) + 0.5 * np.einsum(
    'ikj->kij', np.fabs(EEPS)
)

GG = np.einsum(
    'mij,nkl->mnijkl', GAMMA_inv, GAMMA_inv
)


@provides(IXModel)
class XDomainSinglePoint(BMCSLeafNode):

    U_var_shape = (6,)

    state_var_shape = (1,)

    n_dofs = Property

    def _get_n_dofs(self):
        return 6
    #=========================================================================
    # Serialized domains interface - should be factored out into the base
    #=========================================================================
    dof_offset = Property

    def _get_dof_offset(self):
        return 0

    n_active_elems = Property

    def _get_n_active_elems(self):
        return 1

    def set_next(self, next_):
        pass

    def set_prev(self, prev):
        pass

    #=========================================================================
    # Domain mappings
    #=========================================================================
    def map_U_to_field(self, eps_eng):
        return np.einsum(
            # 'kij,...k->...ij', GAMMA, eps_eng
        'kij,...k->...ij', GAMMA_inv, eps_eng
        )[np.newaxis, ...]

    def map_field_to_F(self, eps_tns):
        dof_E = np.arange(6, dtype=np.int_)
        return dof_E, np.einsum(
            'kij,...ij->...k', GAMMA_inv, eps_tns
        )

    def map_field_to_K(self, tns4):
        K_mij = np.einsum(
            'mnijkl,...ijkl->...mn', GG, tns4
        )
        dof_Ei = np.arange(6, dtype=np.int_)[np.newaxis, ...]
        return SysMtxArray(mtx_arr=K_mij, dof_map_arr=dof_Ei)



#-------------------------------------------------------------------------------
#
# Copyright (c) 2009, IMB, RWTH Aachen.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in simvisage/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.simvisage.com/licenses/BSD.txt
#
# Thanks for using Simvisage open source!
#
# Created on Aug 19, 2009 by: rch

from .mats3D_cmdm import \
    MATS3DMicroplaneDamage, PhiFnStrainSoftening, PhiFnStrainHardening

from ibvpy.tmodel.mats3D.__test__ import TestMATS3D

class TestMATS3DMicroplaneDamageCompliance( TestMATS3D ):
    '''Test the elementary load cases of 3D material models.
    '''

    def _mats_eval_default(self):
        return MATS3DMicroplaneDamage( E = 34000, nu = 0.25,
                                       model_version = 'compliance',
                                       phi_fn = PhiFnStrainHardening() )
     
    def test_symmetry_tension(self):
        '''
        Symmetry of the response ... tensile loading.
        '''
        self.assert_symmetry_on_cube_with_clamped_face( load_dirs = [0,1,2], 
                                                        load = 0.001 )
                        
    def test_symmetry_bending(self):
        '''
        Symmetry of the response ... out of plane load.
        '''
        self.assert_symmetry_on_cube_with_clamped_face( load_dirs = [1,2,0],
                                                        load = 0.00016 )
        return

    def test_stress_value(self):
        '''Assert a correct value of resulting stress a material point.
        '''
        self.assert_stress_value( [ 2.9990935740536142, 0, 0, 0, 0, 0, 0, 0, 0 ],
                                    n_steps = 1, load = 0.0001 )

    def test_total_fracture_energy_value(self):
        '''Assert a correct value of resulting elastic energy in the unit domain
        '''
        self.assert_total_energy_value( 0.00078899038265171733,
                                        ivar = 'fracture_energy',
                                        n_steps = 5, load = 0.01 )
                
class TestMATS3DMicroplaneDamageStiffness( TestMATS3D ):
    '''Test the elementary load cases of 3D material models.
    '''

    def _mats_eval_default(self):
        return MATS3DMicroplaneDamage( E = 34000, nu = 0.25,
                                       model_version = 'stiffness',
                                       phi_fn = PhiFnStrainSoftening(
                                                                  G_f = 0.001117,
                                                                  f_t = 2.8968,
                                                                  md = 0.0,
                                                                  h = 1.0 )
                                        )
     
    def test_symmetry_tension(self):
        '''
        Symmetry of the response ... tensile loading.
        '''
        self.assert_symmetry_on_cube_with_clamped_face( load_dirs = [0,1,2], 
                                                        load = 0.001 )
        
                
    def test_symmetry_bending(self):
        '''
        Symmetry of the response ... out of plane load.
        '''
        self.assert_symmetry_on_cube_with_clamped_face( load_dirs = [1,2,0],
                                                        load = 0.00016 )
        return

    def test_stress_value(self):
        '''Assert a correct value of resulting stress a material point.
        '''
        self.assert_stress_value( [ 2.7904899530421789, 0, 0, 0, 0, 0, 0, 0, 0 ],
                                    n_steps = 1, load = 0.0001 )
        
    def test_total_fracture_energy_value(self):
        '''Assert a correct value of resulting fracture energy in the unit domain
        '''
        self.assert_total_energy_value( 0.0011078180050320837,
                                        ivar = 'fracture_energy',
                                        n_steps = 5, load = 0.01 )
    
if __name__ == "__main__":
    import unittest
    import sys;sys.argv = ['', 'TestMATS3DMicroplaneDamageCompliance.test_total_fracture_energy_value']         
    unittest.main()
    



import wx

import matplotlib
from matplotlib.backends.backend_wx import NavigationToolbar2Wx, StatusBarWx
from matplotlib.backends.backend_wxagg import FigureCanvasWxAgg as FigureCanvas
from matplotlib.figure import Figure
from traits.api import Instance
from traitsui.wx.basic_editor_factory import BasicEditorFactory
from traitsui.wx.editor import Editor


# We want matplotlib to use a wxPython backend
matplotlib.use('WXAgg')


class _MPLFigureEditor(Editor):

    scrollable = True

    def init(self, parent):
        self.control = self._create_canvas(parent)
        self.object.on_trait_change(self.update_editor, 'data_changed')
        self.set_tooltip()

    def update_editor(self):
        figure = self.value
        figure.canvas.mpl_connect('key_press_event', self.key_press_callback)
        figure.canvas.draw()

    def _create_canvas(self, parent):
        """ Create the MPL canvas. """
        # The panel lets us add additional controls.
        panel = wx.Panel(parent, -1, style=wx.CLIP_CHILDREN)
        sizer = wx.BoxSizer(wx.VERTICAL)
        panel.SetSizer(sizer)

        # matplotlib commands to create a canvas
        mpl_control = FigureCanvas(panel, 1, self.value)
        statbar = StatusBarWx(panel)
        toolbar = NavigationToolbar2Wx(mpl_control)
        toolbar.set_status_bar(statbar)

        sizer.Add(toolbar, 0, wx.EXPAND)
        sizer.Add(mpl_control, 1, wx.EXPAND)

        sizer.Add(toolbar.statbar, 0, wx.EXPAND)
        self.value.canvas.SetMinSize((100, 100))
        return panel

    # TODO: generalize
    def key_press_callback(self, event):
        'whenever a key is pressed'
        figure = self.value
        if not event.inaxes:
            return
        if event.key == 'k':
            if figure.axes[0].get_xscale() == 'log':
                figure.axes[0].set_xscale('linear')
                figure.canvas.draw()
            else:
                figure.axes[0].set_xscale('log')
                figure.canvas.draw()

        if event.key == 'l':
            if figure.axes[0].get_yscale() == 'log':
                figure.axes[0].set_yscale('linear')
                figure.canvas.draw()
            else:
                figure.axes[0].set_yscale('log')
                figure.canvas.draw()


class MPLFigureEditor(BasicEditorFactory):

    klass = _MPLFigureEditor




'''
Created on May 25, 2009

@author: jakub
'''
from scipy.optimize import \
    brentq
from traits.api import \
    Array, Int, Property, cached_property, \
    HasTraits, provides, WeakRef, Float, \
    Str, Tuple
import numpy as np
from .i_fe_grid_slice import \
    IFEGridSlice


@provides(IFEGridSlice)
class FEGridLevelSetSlice(HasTraits):

    # Link to the source grid to slice
    #
    fe_grid = WeakRef('ibvpy.mesh.fe_grid.FEGrid')

    # Specification of further methods to be provided by the fe_grid_cell
    #
    # @todo -
    # 1) describe the functionality of the methods
    # 2) decide where to implement the details of the method
    #    Generally, the lookup for elements should be directed
    #    to the geo_grid, while the lookup for individual degrees
    #    of freedom should go into the dof_grid
    #
    ls_function_eval = Str

    ls_spec_list = Property(Tuple, depends_on='ls_function_eval')

    @cached_property
    def _get_ls_spec_list(self):
        fn_list = self.ls_function_eval.split(';')
        fns = []
        lim = []
        for fn in fn_list:
            try:
                afn, limits = fn.split('@')
            except:  # convenience - of there is just one fn, user does not have to add region
                afn = fn
                limits = '1'
            fns.append(afn.upper())  # transform to uppercase
            lim.append(limits)  # limits.upper() )
        return fns, lim

    ls_fns = Property

    def _get_ls_fns(self):
        return self.ls_spec_list[0]

    ls_lims = Property

    def _get_ls_lims(self):
        return self.ls_spec_list[1]

    def ls_function(self, X=None, Y=None, Z=None):
        '''
        get_value() evaluates the value for all functions as defined in set_fns, 
        will be 0 if grid point is not defined in any function
        '''
        val = 0.0
        for fn, limit in zip(self.ls_fns, self.ls_lims):
            if eval(limit):
                val = eval(fn)
        return val

    def ls_mask_function(self, X=None, Y=None, Z=None):
        '''Return True if the point is within one of the ranges of the 
        supplied ls functions.
        '''
        masked = True
        for limit in self.ls_lims:
            if eval(limit):
                if masked == False:
                    raise ValueError(
                        'overlapping level set limits - two level sets defined')
                masked = False
        return masked

    #
    tip_elems = Property(Array(int))

    def _get_tip_elems(self):
        '''Get the elements at the tip of the level set function

        The current implementation does not consider the association
        between the ls_mask function and the ls_function.

        This is probably not correct.

        The tip elems should be removed from the level set alone.

        '''
        ls_mask = self.fe_grid.get_ls_mask(self.ls_mask_function)
        # get the nodes having True value

        masked_nodes = np.argwhere(ls_mask)
        # find out the element associated with these nodes

        neighbour_idx = np.array([[-1, -1],
                                  [-1, 0],
                                  [0, -1],
                                  [0, 0]], dtype=int)

        # get for each node the neighbors
        e_arr = masked_nodes[:, None, :] + neighbour_idx[None, :, :]

        # get for each node the corresponding neighbors
        e_flattened = [e_arr[:, :, idx].flatten()
                       for idx in range(e_arr.shape[2])]
        e_masked = np.c_[tuple(e_flattened)]

        # find the intersection with the level set
        intersected_elem_offsets = self.fe_grid.get_intersected_elems(
            self.ls_function)

        # get the indices of the intersected by the level set
        # this should be done more efficiently
        e_ls = np.array([self.fe_grid.geo_grid.cell_grid.get_cell_idx(offset)
                         for offset in intersected_elem_offsets], dtype=int)

        # make a broadcasted comparison between the masked elements
        # (elements containing a node outside outside the level set domain)
        # and elements intersected by the level set.
        #
        broadcasted_compare = (e_ls == e_masked[:, None, :])

        # along the axis 2 - there is a comparison of the coordinates
        # - if all are True, then all coordinates are equal - this
        # is detected by the boolean product (logical and along axis 2)
        #
        logical_and_along_ax2 = broadcasted_compare.prod(axis=2, dtype=bool)

        # along the axis 0 - there is a each-with-each comparison
        # - if one of the entries is True, then the second list contains
        #   the same entry as well. This is detected by the sum operator
        #  (logical or along axis 0)
        logical_or_along_ax0 = logical_and_along_ax2.sum(axis=0, dtype=bool)

        #res = ( e_ls == e_masked[:, None, :] ).prod( axis = 2, dtype = bool ).sum( axis = 0, dtype = bool )

        return e_ls[logical_or_along_ax0]

    elems = Property(Array(int))

    def _get_elems(self):
        # 1. for each ls_function identify the intersected elements
        # 2. choose those lying within the ls_domain - including the tip
        # elements.
        return self.fe_grid.get_intersected_elems(self.ls_function,
                                                  self.ls_mask_function)

    neg_elems = Property(Array(Int))

    def _get_neg_elems(self):
        return self.fe_grid.get_negative_elems(self.ls_function)

    pos_elems = Property(Array(Int))

    def _get_pos_elems(self):
        # return self.fe_grid.get_positive_elems()
        raise NotImplementedError

    dof_nodes_values = Property(Array(Float))

    @cached_property
    def _get_dof_nodes_values(self):
        #nodes = self.dof_grid.cell_node_map[ self.get_intersected_elems() ]
        # print "map ", nodes
        #coords = self.dof_grid.get_cell_points(self.get_intersected_elems())
        coords = self.dof_X

        # @todo - make the dimensional dependency
        #
        ls_fn = np.frompyfunc(self.ls_function, 2, 1)
        o_shape = coords.shape  # save shape for later
        if coords.shape[2] == 1:
            X = coords.reshape(o_shape[0] * o_shape[1], o_shape[2]).T
            Y = np.zeros_like(X)
        elif coords.shape[2] == 2:
            X, Y = coords.reshape(o_shape[0] * o_shape[1], o_shape[2]).T
        values = ls_fn(X, Y)
        return values.reshape(o_shape[0], o_shape[1])

    geo_X = Property

    def _get_geo_X(self):
        '''Get global coordinates geometrical points of the intersected elements.
        '''
        return self.fe_grid.geo_grid.get_cell_point_X_arr(self.elems)

    geo_x = Property

    def _get_geo_x(self):
        return self.fe_grid.geo_grid.get_cell_point_x_arr(self.elems)

    dof_X = Property

    def _get_dof_X(self):
        '''Get global coordinates field points of the intersected elements.
        '''
        return self.fe_grid.dof_grid.get_cell_point_X_arr(self.elems)

    r_i = Property(Array(float))  # 2 points/element

    @cached_property
    def _get_r_i(self):
        '''
        Return local coordinates of the intersection points
        Assuming rectangular parametric coordinates
        Works for 1D and 2D
        '''
        i_elements = self.fe_grid.get_intersected_elems(self.ls_function)
        el_pnts = []
        for elem in i_elements:
            inter_pts = []
            # X_mtx = self.elements[elem].get_X_mtx() # skips deactivated
            # elements
            X_mtx = self.fe_grid.geo_grid.get_cell_point_X_arr(elem)
            dim = X_mtx.shape[1]  # TODO:merge 1 and 2d
            if dim == 1:
                r_coord = self._get_intersect_pt(self.ls_fn_r, (0., X_mtx))
                if r_coord != None:
                    inter_pts.append([r_coord])
            elif dim == 2:
                for c_coord in [-1., 1.]:
                    args = (c_coord, X_mtx)
                    s_coord = self._get_intersect_pt(self.ls_fn_s, args)
                    r_coord = self._get_intersect_pt(self.ls_fn_r, args)
                    if s_coord != None:
                        inter_pts.append([c_coord, s_coord])
                    if r_coord != None:
                        inter_pts.append([r_coord, c_coord])
            elif dim == 3:
                raise NotImplementedError('not available for 3D yet')
            el_pnts.append(inter_pts)
        return np.array(el_pnts)

    dofs = Property

    def _get_dofs(self):
        '''Get number of affected DOFs

        @todo - reshape dofs so that they are of dimension
        ( elem, node, dim )
        '''
        dofs = []
        for elem in self.elems:
            dof_map = self.fe_grid.dof_grid.get_cell_dofs(elem).flatten()
            dofs.append(dof_map)
        return np.array(dofs)

    i_neg_dofs = Property

    def _get_i_neg_dofs(self):
        '''Get intersected dofs with negative value of the level set at their position.
        '''
        shape = self.dofs.shape
        n_node_dofs = shape[1] / self.dof_X.shape[1]
        return self.dofs.reshape(shape[0],
                                 self.dof_X.shape[1],
                                 n_node_dofs)[self.dof_nodes_values < 0.]

    neg_dofs = Property

    def _get_neg_dofs(self):
        '''Get dofs with negative value of the level set at their position.
        '''
        dofs = []
        for elem in self.neg_elems:
            dof_map = self.fe_grid.dof_grid.get_cell_dofs(elem).flatten()
            dofs.append(dof_map)
        if dofs == []:
            return np.array(dofs)
        return np.hstack(dofs)

    def _get_intersect_pt(self, fn, args):
        try:
            return brentq(fn, -1, 1, args=args)
        except ValueError:
            return

    def ls_fn_r(self, r, s, X_mtx):  # TODO:dimensionless treatment
        X_pnt = self.fe_grid.fets_eval.map_r2X([r, s], X_mtx)
        if X_pnt.shape[0] == 1:
            Y = 0.
        else:
            Y = X_pnt[1]
        return self.ls_function(X_pnt[0], Y)

    def ls_fn_s(self, s, r, X_mtx):
        X_pnt = self.fe_grid.fets_eval.map_r2X([r, s], X_mtx)
        return self.ls_function(X_pnt[0], X_pnt[1])

    def ls_fn_X(self, X, Y):
        return self.ls_function(X, Y)


if __name__ == '__main__':
    from ibvpy.mesh.fe_grid import FEGrid
    from ibvpy.fets.fets2D.fets2D4q import FETS2D4Q
    from ibvpy.tmodel.mats2D.mats2D_elastic.mats2D_elastic import MATS2DElastic

    fets_eval = FETS2D4Q(mats_eval=MATS2DElastic())
    fe_grid1 = FEGrid(coord_max=(3., 2.),
                      shape=(3, 2),
                      fets_eval=fets_eval)

    def ls_fn(X, Y):
        # print "point ", X," ",Y
        #print "ls value ", X - 1.5#
        return X - 1.5  # Y-0.2*X#temp

    print('fe_grid elemes')
    print(fe_grid1.geo_grid.cell_grid.cell_idx_grid)

    print('vertex_X_grid')
    print(fe_grid1.geo_grid.cell_grid.vertex_X_grid)

    fe_slice = fe_grid1['X - 1.5']
    print('slice ')
    print(fe_slice)
    print('elems')
    print(fe_slice.elems)
    print('node values')
    print(fe_slice.dof_nodes_values)
    print('intersecting points')
    print(fe_slice.r_i)
    print('dofs')
    print(fe_slice.dofs)
    print('geo coords')
    print(fe_slice.geo_X)
    print('dof coords')
    print(fe_slice.dof_X)
    print('neg i dofs')
    print(fe_slice.i_neg_dofs)
    print('neg dofs')
    print(fe_slice.neg_dofs)




from ibvpy.mathkit.linalg.sys_mtx_assembly import SysMtxArray
from ibvpy.sim.i_xmodel import IXModel
from traits.api import \
    provides, Int, \
    Array, Property, cached_property
import numpy as np
from .xdomain_fe_grid import XDomainFEGrid


@provides(IXModel)
class XDomainFEInterface1D(XDomainFEGrid):

    A = Property()
    '''array containing the A_m, L_b, A_f
    '''

    def _get_A(self):
        return np.array([self.fets.A_m,
                         self.fets.P_b,
                         self.fets.A_f])

    DELTA_p = Array()

    def _DELTA_p_default(self):
        return np.array([-1, 1], np.float64)

    o_Epia = Property(depends_on='+input')
    '''For a given element, layer, node number and dimension
    return the dof number
    '''
    @cached_property
    def _get_o_Epia(self):
        dof_Eipd = self.mesh.dof_grid.cell_dof_map[..., np.newaxis]
        return np.einsum('Eipd->Epid', dof_Eipd)

    dim_u = Int(2)

    B_Eimabc = Property(depends_on='MESH,GEO,CS,FE')
    '''Kinematic mapping between displacements and strains in every
    integration point.
    '''
    @cached_property
    def _get_B_Eimabc(self):

        fets_eval = self.fets
        mesh = self.mesh
        n_s = 3

        n_dof_r = fets_eval.n_dof_r
        n_nodal_dofs = fets_eval.n_nodal_dofs

        n_ip = fets_eval.n_gp
        n_e = mesh.n_active_elems
        #[ d, i]
        r_ip = fets_eval.ip_coords[:, :-2].T
        # [ d, n ]
        geo_r = fets_eval.geo_r.T
        # [ d, n, i ]
        dNr_geo = geo_r[:, :, None] * np.array([1, 1]) * 0.5
        # [ i, n, d ]
        dNr_geo = np.einsum('dni->ind', dNr_geo)

        inv_J_Emar = np.linalg.inv(self.J_Emar)

        # shape function for the unknowns
        # [ d, n, i]
        Nr = 0.5 * (1. + geo_r[:, :, None] * r_ip[None, :])
        dNr = 0.5 * geo_r[:, :, None] * np.array([1, 1])

        # [ i, n, d ]
        Nr = np.einsum('dni->ind', Nr)
        dNr = np.einsum('dni->ind', dNr)
        Nx = Nr
        # [ n_e, n_ip, n_dof_r, n_dim_dof ]
        dNx = np.einsum('Eidf,inf->Eind', inv_J_Emar, dNr)

        B = np.zeros((n_e, n_ip, n_dof_r, n_s, n_nodal_dofs), dtype='f')
        B_N_n_rows, B_N_n_cols, N_idx = [1, 1], [0, 1], [0, 0]
        B_dN_n_rows, B_dN_n_cols, dN_idx = [0, 2], [0, 1], [0, 0]
        B_factors = np.array([-1, 1], dtype='float64')
        B[:, :, :, B_N_n_rows, B_N_n_cols] = (B_factors[None, None, :] *
                                              Nx[:, :, N_idx])
        B[:, :, :, B_dN_n_rows, B_dN_n_cols] = dNx[:, :, :, dN_idx]
        return B

    def _get_BB_Emicjdabef(self):
        return np.einsum(
            's, ...Emisd,...Emjtf, Em, m->...Emidjfst',
            self.A, self.B_Eimabc, self.B_Eimabc, self.det_J_Em, self.fets.w_m
        )

    def map_U_to_field(self, U):
        U_Eia = U[...,self.o_Eia]
        s_Emr = np.einsum(
            '...Emisd,...Eid->...Ems',
            self.B_Eimabc, U_Eia
        )
        return s_Emr

    def map_field_to_K(self, D_Emabef):
        K_Eicjd = self.integ_factor * np.einsum(
            '...Emirjsab,...Emab->Eirjs',
            self.BB_Emicjdabef, D_Emabef
        )
        _, _, n_i, _, n_a = self.B_Eimabc.shape
        n_o = n_i * n_a
        K_Eij = K_Eicjd.reshape(-1, n_o, n_o)
        o_Ei = self.o_Eia.reshape(-1, n_o)
        return SysMtxArray(mtx_arr=K_Eij, dof_map_arr=o_Ei)

    def map_field_to_F(self, sig_Emab):
        _, _, n_i, _, n_a = self.B_Eimabc.shape
        f_Eic = self.integ_factor * np.einsum(
            's,m,Emisd,...Ems,Em->...Eid',
            self.A, self.fets.w_m, self.B_Eimabc, sig_Emab, self.det_J_Em
        )
        n_o = n_i * n_a
        f_Ei = f_Eic.reshape(-1, n_o)
        o_E = self.o_Eia.reshape(-1, n_o)
        return o_E.flatten(), f_Ei.flatten()




from .plot2d import Vis2D, Viz2D, Viz2DTimeFunction
from .ui import BMCSLeafNode, BMCSRootNode, BMCSTreeNode, itags_str
# from .window import BMCSWindow, PlotPerspective
from .window.i_bmcs_model import IBMCSModel



from ibvpy.sim.i_xmodel import IXModel
from traits.api import \
    Property, cached_property, \
    provides, \
    Array
import bmcs_utils.api as bu
import numpy as np

from .xdomain_fe_grid import XDomainFEGrid


@provides(IXModel)
class XDomainFEGridAxiSym(XDomainFEGrid):

    vtk_expand_operator = Array(np.float64)

    def _vtk_expand_operator_default(self):
        return np.identity(3)

    Diff0_factor = bu.Float(1, BC=True)

    Diff0_abc = Array(np.float64)

    def _Diff0_abc_default(self):
        D3D_33 = np.array([[0, 0, 0],
                           [0, 0, 0],
                           [0, 0, 1]], np.float64)
        D2D_22 = np.array([[0, 0],
                           [0, 1]], np.float64)
        return np.einsum('ab,cc->abc', D3D_33, D2D_22)

    Diff1_abcd = Array(np.float64)

    def _Diff1_abcd_default(self):
        delta = np.vstack([np.identity(2),
                           np.zeros((1, 2), dtype=np.float64)])
        return 0.5 * (
            np.einsum('ac,bd->abcd', delta, delta) +
            np.einsum('ad,bc->abcd', delta, delta)
        )

    det_J_Em = Property(depends_on='MESH,GEO,CS,FE')
    '''Jacobi matrix in integration points
    '''
    @cached_property
    def _get_det_J_Em(self):
        r_Em = np.einsum(
            'im,Eic->Emc',
            self.fets.N_im, self.x_Eia
        )[..., 1]
        return r_Em * np.linalg.det(self.J_Emar)

    B0_Eimabc = Property(depends_on='+input')

    @cached_property
    def _get_B0_Eimabc(self):
        x_Eia = self.x_Eia
        r_Em = np.einsum(
            'im,Eic->Emc',
            self.fets.N_im, x_Eia
        )[..., 1]
        B0_Eimabc = np.einsum(
            'abc,im, Em->Eimabc',
            self.Diff0_abc, self.fets.N_im, 1. / r_Em
        )
        return B0_Eimabc

    B_Eimabc = Property(depends_on='MESH,GEO,CS,FE')
    '''Kinematic mapping between displacements and strains in every
    integration point.
    '''
    @cached_property
    def _get_B_Eimabc(self):
        return self.B1_Eimabc + self.B0_Eimabc



#-------------------------------------------------------------------------
#
# Copyright (c) 2009, IMB, RWTH Aachen.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in simvisage/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.simvisage.com/licenses/BSD.txt
#
# Thanks for using Simvisage open source!
#
# Created on Sep 4, 2009 by: rch

from math import copysign, sin

from traits.api import \
    Trait,    \
    Dict

from ibvpy.tmodel.mats1D.mats1D_eval import MATS1DEval
import numpy as np
import bmcs_utils.api as bu

# from dacwt import DAC
def sign(val):
    return copysign(1, val)


#---------------------------------------------------------------------------
# Material time-step-evaluator for Scalar-Damage-Model
#---------------------------------------------------------------------------


class MATS1DPlastic(MATS1DEval):

    '''
    Scalar Damage Model.
    '''

    E = bu.Float(1.,  # 34e+3,
              label="E",
              desc="Young's Modulus",
              enter_set=True,
              auto_set=False)

    sigma_y = bu.Float(1.,
                    label="sigma_y",
                    desc="Yield stress",
                    enter_set=True,
                    auto_set=False)

    K_bar = bu.Float(0.1,  # 191e-6,
                  label="K",
                  desc="Plasticity modulus",
                  enter_set=True,
                  auto_set=False)

    H_bar = bu.Float(0.1,  # 191e-6,
                  label="H",
                  desc="Hardening modulus",
                  enter_set=True,
                  auto_set=False)

    #--------------------------------------------------------------------------
    # View specification
    #--------------------------------------------------------------------------

    traits_view = bu.View(
        bu.Item('E'),
        bu.Item('sigma_y'),
        bu.Item('K_bar'),
        bu.Item('H_bar'),
        bu.Item('stiffness'),
    )

    #-------------------------------------------------------------------------
    # Setup for computation within a supplied spatial context
    #-------------------------------------------------------------------------

    def get_state_array_size(self):
        '''
        Give back the nuber of floats to be saved
        @param sctx:spatial context

        eps_p_n - platic strain 
        alpha_n - hardening
        q_n - back stress  

        '''
        return 3

    def new_cntl_var(self):
        return np.zeros(1, np.float64)

    def new_resp_var(self):
        return np.zeros(1, np.float64)

    #-------------------------------------------------------------------------
    # Evaluation - get the corrector and predictor
    #-------------------------------------------------------------------------
    def get_corr_pred(self, sctx, eps_app_eng, d_eps, tn, tn1, eps_avg=None):
        '''
        Corrector predictor computation.
        @param eps_app_eng input variable - engineering strain
        '''
        eps_n1 = float(eps_app_eng)
        E = self.E

        if eps_avg == None:
            eps_avg = eps_n1

        if sctx.update_state_on:
            eps_n = eps_avg - float(d_eps)
            sctx.mats_state_array[:] = self._get_state_variables(sctx, eps_n)

        eps_p_n, q_n, alpha_n = sctx.mats_state_array
        sigma_trial = self.E * (eps_n1 - eps_p_n)
        xi_trial = sigma_trial - q_n
        f_trial = abs(xi_trial) - (self.sigma_y + self.K_bar * alpha_n)

        sig_n1 = np.zeros((1,), dtype='float64')
        D_n1 = np.zeros((1, 1), dtype='float64')
        if f_trial <= 1e-8:
            sig_n1[0] = sigma_trial
            D_n1[0, 0] = E
        else:
            d_gamma = f_trial / (self.E + self.K_bar + self.H_bar)
            sig_n1[0] = sigma_trial - d_gamma * self.E * sign(xi_trial)
            D_n1[0, 0] = (self.E * (self.K_bar + self.H_bar)) / \
                (self.E + self.K_bar + self.H_bar)

        return sig_n1, D_n1

    #--------------------------------------------------------------------------
    # Subsidiary methods realizing configurable features
    #--------------------------------------------------------------------------

    def _get_state_variables(self, sctx, eps_n):

        eps_p_n, q_n, alpha_n = sctx.mats_state_array

        # Get the characteristics of the trial step
        #
        sig_trial = self.E * (eps_n - eps_p_n)
        xi_trial = sig_trial - q_n
        f_trial = abs(xi_trial) - (self.sigma_y + self.K_bar * alpha_n)

        if f_trial > 1e-8:

            #
            # Tha last equilibrated step was inelastic. Here the
            # corresponding state variables must be calculated once
            # again. This might be expensive for 2D and 3D models. Then,
            # some kind of caching should be considered for the state
            # variables determined during iteration. In particular, the
            # computation of d_gamma should be outsourced into a separate
            # method that can in general perform an iterative computation.
            #
            d_gamma = f_trial / (self.E + self.K_bar + self.H_bar)
            eps_p_n += d_gamma * sign(xi_trial)
            q_n += d_gamma * self.H_bar * sign(xi_trial)
            alpha_n += d_gamma

        newarr = np.array([eps_p_n, q_n, alpha_n], dtype='float64')

        return newarr

    #-----------------------------------------------------------
    # Response trace evaluators
    #--------------------------------------------------------------------------
    def get_eps_p(self, sctx, eps_app_eng):
        return np.array([sctx.mats_state_array[0]])

    def get_q(self, sctx, eps_app_eng):
        return np.array([sctx.mats_state_array[1]])

    def get_alpha(self, sctx, eps_app_eng):
        return np.array([sctx.mats_state_array[2]])

    # Declare and fill-in the rte_dict - it is used by the clients to
    # assemble all the available time-steppers.
    #
    rte_dict = Trait(Dict)

    def _rte_dict_default(self):
        return {'sig_app': self.get_sig_app,
                'eps_app': self.get_eps_app,
                'eps_p': self.get_eps_p,
                'q': self.get_q,
                'alpha': self.get_alpha}




from traits.api import \
    HasTraits, List, Array, Property, cached_property, \
    Instance, Trait, Button, on_trait_change, Tuple, \
    Int, Float

from ibvpy.mesh.sdomain import \
    SDomain

from numpy import \
    array, unique, min, max, mgrid, ogrid, c_, repeat, ix_, \
    arange, ones, zeros, multiply, sort, allclose, index_exp

# from ibvpy.view.plot3d.mayavi_util.pipelines import \
#     MVPolyData, MVPointLabels

class CellSpec( HasTraits ):
    '''
    '''
    node_coords = Array( float, value = [[-1, -1, 0],
                         [ 1, -1, 0],
                         [ 1, 1, 0],
                         [ 1, 1, 1],
                         [-1, 1, -1],
                         [-1 / 2., 1, 0],
                         [ 0.  , 1, 0],
                         [ 1 / 2., 1, 0]] )


#    xnode_coords = List( [[-1,0,-1],
#                          [ 1,0,-1],
#                          [ 1,0, 1],
#                          [-1,0, 1]] )
#
#    node_coords = List( [[-1,-1,-1],
#                         [ 1, 0, 0],
#                         [ 1, 1, 1],
#                         ] )

    _node_array = Property( Array( 'float64' ), depends_on = 'node_coords' )
    @cached_property
    def _get__node_array( self ):
        '''Get the node array as float64
        '''
        # check that the nodes are equidistant
        return array( self.node_coords, 'float64' )

    n_dims = Property( depends_on = 'node_coords' )
    @cached_property
    def _get_n_dims( self ):
        '''Get the number of dimension of the cell
        '''
        return self._node_array.shape[1]

    def get_cell_shape( self ):
        '''Get the shape of the cell grid.
        '''
        cell_shape = ones( 3, dtype = int )
        ndims = self.n_dims
        narray = self._node_array
        cell_shape[0:ndims] = array( [len( unique( narray[:, i] ) )
                                      for i in range( ndims ) ], dtype = int )
        cell_shape = array( [len( unique( narray[:, i] ) )
                                      for i in range( ndims ) ], dtype = int )
        return cell_shape

    def get_cell_slices( self ):
        '''Get slices for the generation of the cell grid.
        '''
        ndims = self.n_dims
        narray = self._node_array
        return tuple( [ slice( 
                              min( narray[:, i] ),
                              max( narray[:, i] ),
                              complex( 0, len( unique( narray[:, i] ) ) ),
                              )
                        for i in range( ndims ) ] )

    #-------------------------------------------------------------------
    # Visualization-related specification
    #-------------------------------------------------------------------
    cell_lines = Array( int, value = [[0, 1], [1, 2], [2, 0]] )
    cell_faces = Array( int, value = [[0, 1, 2]] )


class GridCell( SDomain ):
    '''
    A single mgrid cell for geometrical representation of the domain.
    
    Based on the grid_cell_spec attribute, 
    the node distribution is determined.
    
    '''
    # Everything depends on the grid_cell_specification
    #
    grid_cell_spec = Instance( CellSpec )
    def _grid_cell_spec_default( self ):
        return CellSpec()

    # Generated grid cell coordinates as they come from mgrid.
    # The dimensionality of the mgrid comes from the 
    # grid_cell_spec_attribute
    #
    grid_cell_coords = Property( depends_on = 'grid_cell_spec' )
    @cached_property
    def _get_grid_cell_coords( self ):
        grid_cell = mgrid[ self.grid_cell_spec.get_cell_slices() ]
        return c_[ tuple( [ x.flatten() for x in grid_cell ] ) ]

    n_nodes = Property( depends_on = 'grid_cell_spec' )
    @cached_property
    def _get_n_nodes( self ):
        '''Return the number of all nodes within the cell.
        '''
        return self.grid_cell_coords.shape[0]

    # Node map lists the active nodes within the grid cell
    # in the specified order
    #
    node_map = Property( Array( int ), depends_on = 'grid_cell_spec' )
    @cached_property
    def _get_node_map( self ):
        n_map = []
        for node in self.grid_cell_spec._node_array:
            for idx, grid_cell_node in enumerate( self.grid_cell_coords ):
                if allclose( node , grid_cell_node , atol = 1.0e-3 ):
                    n_map.append( idx )
                    continue
        return array( n_map, int )

    # #-----------------------------------------------------------------
    # # Visualization related methods
    # #-----------------------------------------------------------------
    # mvp_mgrid_ngeo_labels = Trait( MVPointLabels )
    # def _mvp_mgrid_ngeo_labels_default( self ):
    #     return MVPointLabels( name = 'Geo node numbers',
    #                               points = self._get_points,
    #                               scalars = self._get_node_distribution )

    refresh_button = Button( 'Draw' )
    @on_trait_change( 'refresh_button' )
    def redraw( self ):
        '''
        '''
        self.mvp_mgrid_ngeo_labels.redraw( 'label_scalars' )

    def _get_points( self ):
        points = self.grid_cell_coords[ ix_( self.node_map ) ]
        shape = points.shape
        if shape[1] < 3:
            _points = zeros( ( shape[0], 3 ), dtype = float )
            _points[:, 0:shape[1]] = points
            return _points
        else:
            return points

    def _get_node_distribution( self ):
        #return arange(len(self.node_map))
        n_points = self.grid_cell_coords.shape[0]
        full_node_map = ones( n_points, dtype = float ) * -1.
        full_node_map[ ix_( self.node_map ) ] = arange( len( self.node_map ) )
        return full_node_map

    def __getitem__( self, idx ):
        # construct the full boolean map of the grid cell'        
        node_bool_map = repeat( False, self.n_nodes ).reshape( self.grid_cell_spec.get_cell_shape() )
        # put true at the sliced positions         
        node_bool_map[idx] = True
        # extract the used nodes using the node map
        node_selection = node_bool_map.flatten()[ self.node_map ]
        return node_selection





import traits.api as tr
import numpy as np

class MATS2DCalibrationMixinGf(tr.HasTraits):
    """Calibrate the values of the material parameters to render the desired stiffness
    strength and fracture energy"""

    def get_sig_eps(self):
        eps_max = self.eps_max
        n_eps = 5000
        eps11_range = np.linspace(1e-9, eps_max, n_eps)
        eps_range = np.zeros((len(eps11_range), 2, 2))
        eps_range[:, 1, 1] = eps11_range
        state_vars = {var: np.zeros((len(eps11_range),) + shape)
                      for var, shape in self.state_var_shapes.items()
                      }
        sig_range, _ = self.get_corr_pred(eps_range, 1, **state_vars)
        sig11_range = sig_range[:, 1, 1]
        eps11_range = eps_range[:, 1, 1]
        argmax_i = np.argmax(sig11_range)
        max_sig, argmax_eps = sig11_range[argmax_i], eps11_range[argmax_i]
        G_f = np.trapz(sig11_range, eps11_range)
        return G_f, max_sig, argmax_eps



#-------------------------------------------------------------------------------
#
# Copyright (c) 2009, IMB, RWTH Aachen.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in simvisage/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.simvisage.com/licenses/BSD.txt
#
# Thanks for using Simvisage open source!
#
# Created on Oct 26, 2010 by: rch

from ibvpy.api import\
     BCSlice, TStepper as TS, TLoop, TLine, RTDofGraph
from ibvpy.rtrace.rt_domain_list_field import RTraceDomainListField
from ibvpy.mesh.fe_grid import FEGrid
from ibvpy.mesh.fe_refinement_grid import FERefinementGrid
from ibvpy.mesh.fe_domain import FEDomain

from ibvpy.fets.fets1D.fets1D2l import FETS1D2L
from ibvpy.tmodel.mats1D import MATS1DElastic

if __name__ == '__main__':

    fets_eval = FETS1D2L( mats_eval = MATS1DElastic( E = 1 ) )

    # Discretization
    fe_domain = FEDomain()

    fe_patch_left = FERefinementGrid( name = 'left',
                                     fets_eval = fets_eval,
                                     domain = fe_domain )

    fe_grid_left = FEGrid( level = fe_patch_left,
                      coord_min = ( 0., ),
                      coord_max = ( 1., ),
                      shape = ( 1, ),
                      fets_eval = fets_eval )

    fe_patch_right = FERefinementGrid( name = 'refinement grid',
                                       fets_eval = fets_eval,
                                       domain = fe_domain )

    fe_grid_right = FEGrid( level = fe_patch_right,
                            coord_min = ( 2., ),
                            coord_max = ( 3., ),
                            shape = ( 1, ),
                            fets_eval = fets_eval )

    dofs_left = fe_grid_left[-1, -1].dofs[0, :, 0]
    dofs_right = fe_grid_right[0, 0].dofs[0, :, 0]
    spring_arr = FESpringArray( domain = fe_domain,
                                dofs_1 = dofs_left,
                                dofs_2 = dofs_right, k_value = 1.0 )

    ts = TS( dof_resultants = True,
             sdomain = fe_domain,
             bcond_list = [BCSlice( var = 'u', value = 1., dims = [0],
                                       slice = fe_grid_right[ -1, -1 ] ),
                           BCSlice( var = 'u', value = 0., dims = [0],
                                       slice = fe_grid_left[ 0, 0 ] )
                                       ],
             rtrace_list = [  ]
                )

    # Add the time-loop control
    tloop = TLoop( tstepper = ts,
                   tline = TLine( min = 0.0, step = 1, max = 1.0 ) )

    u = tloop.eval()
    print(u)

    import pylab as p
    spring_arr.plot_spring_forces( u, p )
    p.show()

#    from ibvpy.plugins.ibvpy_app import IBVPyApp
#    ibvpy_app = IBVPyApp( ibv_resource = tloop )
#    ibvpy_app.main()





from ibvpy.tmodel.mats3D.mats3D_eval import MATS3DEval
from ibvpy.util.traits.either_type import \
    EitherType
import bmcs_utils.api as bu
import numpy as np
import traits.api as tr

from .vstrain_norm2d import Rankine

#---------------------------------------------------------------------------
# Material time-step-evaluator for Scalar-Damage-Model
#---------------------------------------------------------------------------
class MATS3DScalarDamage(MATS3DEval):
    r'''
    Isotropic damage model.
    '''
    node_name = 'Scalar damage'

    stiffness = tr.Enum("secant", "algorithmic",
                        input=True)
    r'''Selector of the stiffness calculation.
    '''
    strain_norm = EitherType(klasses=[Rankine,
                                      ], MAT=True)
    r'''Selector of the strain norm defining the load surface.
    '''

    # TODO - generalize
    # damage_fn = EitherType(klasses=[], MAT=True)
    # r'''Selector of the damage function.
    # '''

    epsilon_0 = bu.Float(5e-2,
                         label="eps_0",
                         desc="Strain at the onset of damage",
                         auto_set=False,
                         MAT=True)
    r'''Damage function parameter - slope of the damage function.
    '''

    epsilon_f = bu.Float(191e-1,
                         label="eps_f",
                         desc="Slope of the damage function",
                         auto_set=False,
                         MAT=True)
    r'''Damage function parameter - slope of the damage function.
    '''

    changed = tr.Event
    r'''This event can be used by the clients to trigger 
    an action upon the completed reconfiguration of 
    the material model
    '''

    state_var_shapes = {'kappa': (),
                        'omega': ()}
    r'''
    Shapes of the state variables
    to be stored in the global array at the level 
    of the domain.
    '''

    def init(self, kappa, omega):
        r'''
        Initialize the state variables.
        '''
        kappa[...] = 0
        omega[...] = 0

    def get_corr_pred(self, eps_Emab_n1, tn1, kappa, omega):
        r'''
        Corrector predictor computation.
        '''
        self.update_state_variables(eps_Emab_n1, kappa, omega)
        phi_Em = (1.0 - omega)
        D_Emabcd = np.einsum(
            '...,abcd->...abcd',
            phi_Em, self.D_abef
        )
        sigma_Emab = np.einsum(
            '...abcd,...cd->...ab',
            D_Emabcd, eps_Emab_n1
        )

        # algorithmic switched off - because the derivative
        # of the strain norm is still not available
        if False:  # algorithmic:
            D_Emabcd_red_I = self._get_D_abcd_alg_reduction(
                kappa[I], eps_Emab_n1[I])
            D_Emabcd[I] -= D_Emabcd_red_I

        return sigma_Emab, D_Emabcd

    def update_state_variables(self, eps_Emab, kappa, omega):
        eps_eq_Em = self.strain_norm.get_eps_eq(eps_Emab, kappa)
        f_trial_Em = eps_eq_Em - self.epsilon_0
        I = np.where(f_trial_Em > 0)
        kappa[I] = eps_eq_Em[I]
        omega[I] = self._get_omega(eps_eq_Em[I])
        return I

    def _get_omega(self, kappa_Em):
        r'''
        Return new value of damage parameter
        @param kappa_Em: maximum strain norm achieved so far 
        '''
        omega_Em = np.zeros_like(kappa_Em)
        epsilon_0 = self.epsilon_0
        epsilon_f = self.epsilon_f
        I = np.where(kappa_Em >= epsilon_0)
        omega_Em[I] = (
            1.0 - (epsilon_0 / kappa_Em[I] *
                   np.exp(-1.0 * (kappa_Em[I] - epsilon_0) /
                          (epsilon_f - epsilon_0))
                   ))
        return omega_Em

    def _get_domega(self, kappa_Em):
        '''
        Return new value of damage parameter derivative
        @param kappa_Em: maximum strain norm achieved so far
        '''
        epsilon_0 = self.epsilon_0
        epsilon_f = self.epsilon_f
        domega_Em = np.zeros_like(kappa_Em)
        I = np.where(kappa_Em >= epsilon_0)
        factor_1 = epsilon_0 / (kappa_Em[I] * kappa_Em[I])
        factor_2 = epsilon_0 / (kappa_Em[I] * (epsilon_f - epsilon_0))
        domega_Em[I] = (
            (factor_1 + factor_2) * np.exp(-(kappa_Em[I] - epsilon_0) /
                                           (epsilon_f - epsilon_0))
        )
        return domega_Em

    def _get_D_abcd_alg_reduction(self, kappa_Em, eps_Emab_n1):
        '''Calculate the stiffness term to be subtracted
        from the secant stiffness to get the algorithmic stiffness.
        '''
        domega_Em = self._get_domega(kappa_Em)
        deps_eq_Emcd = self.strain_norm.get_deps_eq(eps_Emab_n1)
        return np.einsum('...,...cd,abcd,...cd->...abcd',
                         domega_Em, deps_eq_Emcd, self.D_abef, eps_Emab_n1)

    traits_view = bu.View(
        bu.Item('E'),
        bu.Item('nu'),
        bu.Item('epsilon_0'),
        bu.Item('epsilon_f'),
        bu.Item('strain_norm'),
        bu.Item('stiffness', style='custom'),
    )

    # Declare and fill-in the rte_dict - it is used by the clients to
    # assemble all the available time-steppers.
    #
    rte_dict = tr.Trait(tr.Dict)

    def _rte_dict_default(self):
        return {'omega': self.get_omega}





from .mats3D_elastic import MATS3DElastic
from .mats3D_microplane import MATS3DMplCSDEEQ, MATS3DMplCSDODF
from .mats3D_microplane import MATS3DMplDamageEEQ, MATS3DMplDamageODF
from .mats3D_plastic.vmats3D_desmorat import MATS3DDesmorat
from .mats3D_sdamage.vmats3D_sdamage import MATS3DScalarDamage


#from mats3D_cmdm.mats3D_cmdm import MATS3DMicroplaneDamage
#from mats3D_plastic.mats3D_plastic import MATS3DPlastic



'''
Created on Jul 15, 2017

@author: rch
'''
from traits.api import \
    HasStrictTraits, Instance, Str
from ibvpy.view.window import BMCSModel
from ibvpy.view.window.bmcs_viz_sheet import VizSheet



'''
@author: rch
'''

from ibvpy.sim.tline import TLine
from traits.api import \
    Instance, List, Bool, on_trait_change, Str, provides
from ibvpy.view.ui.bmcs_tree_node import \
    BMCSRootNode

import numpy as np


#@provides(ISimulator)
class DELETEBMCSModel(BMCSRootNode):
    '''Base class for models included in the iMCS Tool Suite.
    It implements the state dependencies within the model tree.
    It handles also the communication between the model and
    the user interface in several modes of interaction.

    The scenarios of usage are implemented in 
    bmcs_interaction_patterns
    '''
    title = Str

    desc = Str

    tline = Instance(TLine)

    def _tline_default(self):
        # assign the parameters for solver and loading_scenario
        t_max = 1.0  # self.loading_scenario.t_max
        d_t = 0.1  # self.loading_scenario.d_t
        return TLine(min=0.0, step=d_t, max=t_max,
                     time_change_notifier=self.time_changed,
                     time_range_change_notifier=self.time_range_changed
                     )

    def time_changed(self, time):
        if self.ui != None:
            self.ui.viz_sheet.time_changed(time)

    def time_range_changed(self, tmax):
        self.tline.max = tmax
        if self.ui != None:
            self.ui.viz_sheet.time_range_changed(tmax)

    def set_tmax(self, time):
        self.time_range_changed(time)

    def init(self):
        if self._paused:
            self._paused = False
        if self._restart:
            self.tline.val = self.tline.min
            self.tline.max = 1
            self._restart = False
            self.init_state()
            self.timesteps = []

    def paused(self):
        self.paused_state()
        self._paused = True

    def stop(self):
        self.stop_state()
        self._restart = True

    _paused = Bool(False)
    _restart = Bool(True)

    def run(self):
        '''Run starts or resumes the simulation depending on the 
        pause or restart variables.
        '''
        self.init()
        if self.ui:
            # inform ui that the simulation is running in a thread
            self.ui.start_event = True
            self.ui.running = True
        try:
            # start the calculation
            self.eval()
        except Exception as e:
            if self.ui:
                self.ui.running = False
            raise e  # re-raise exception
        if self.ui:
            # cleanup ui and send the finish event
            self.ui.running = False
            self.ui.finish_event = True

    def init_state(self):
        '''Method called upon start event.
        '''
        pass

    def paused_state(self):
        '''Method called upon the pause event.
        '''
        pass

    def stop_state(self):
        '''Method called upon the stop event.
        '''
        pass

    def eval(self):
        '''Method called upon the run event
        must support the resume calculation.
        '''
        raise NotImplemented

    def add_timestep(self, t):
        self.tline.val = min(t, self.tline.max)
        self.timesteps.append(t)

    timesteps = List()

    def get_time_idx_arr(self, vot):
        '''Get the index corresponding to visual time
        '''
        x = np.array(self.timesteps, dtype=np.float64)
        idx = np.array(np.arange(len(x)), dtype=np.float64)
        t_idx = np.interp(vot, x, idx)
        return np.array(t_idx, np.int_)

    def get_time_idx(self, vot):
        return int(self.get_time_idx_arr(vot))

    @on_trait_change('MAT,ALG,CS,GEO,BC,+BC')
    def signal_reset(self):
        '''Upon the change of the model parameters,
        signal the user interface that further calculation
        does not make sense.
        '''
        if self.ui:
            self.ui.stop()



'''
Created on May 11, 2009

@author: jakub
'''
import unittest
from ibvpy.mesh.fe_domain import FEDomain
from ibvpy.mesh.fe_refinement_grid import FERefinementGrid
from ibvpy.mesh.fe_grid import FEGrid
from ibvpy.fets.fets2D.fets2D4q import FETS2D4Q
from ibvpy.fets.fets3D.fets3D8h import FETS3D8H
from ibvpy.tmodel.mats2D.mats2D_elastic.mats2D_elastic import MATS2DElastic
from numpy import array

class FEDomainDynamicDofMap( unittest.TestCase ):
    '''
    Test class is to be used to test the sequence of changes of the FEDomain
    '''
    def setUp( self ):
        '''
        Construct the FEDomain with two FERefinementGrids (2,2) 
        '''
        self.domain1 = FEDomain()
        self.fets_eval = FETS2D4Q( mats_eval = MATS2DElastic() )
        self.d1 = FERefinementGrid( name = 'd1', domain = self.domain1 )
        g1 = FEGrid( coord_max = ( 1., 1., 0. ),
                     shape = ( 2, 2 ),
                     fets_eval = self.fets_eval,
                     level = self.d1 )
        self.d2 = FERefinementGrid( name = 'd2', domain = self.domain1 )
        g2 = FEGrid( coord_min = ( 1., 0., 0. ),
                     coord_max = ( 2., 1., 0. ),
                     shape = ( 2, 2 ),
                     fets_eval = self.fets_eval,
                     level = self.d2 )

    def test_n_dofs( self ):
        '''Verify the n_dofs (it implicitly verifies the dof_offset)
        '''
        n_dofs = self.domain1.n_dofs
        self.assertEqual( n_dofs, 36 )

    def test_rg_addition( self ):
        '''Check numbering after addition of FERefinementGrid
        Add another FERefinementGrid (2,2) as a child of grid 1
        Check the n_dofs of FEDomain to verify the re-enumeration 
        Check the elem_dof_map of the grid 3. 
        '''
        d3 = FERefinementGrid( name = 'd3', parent = self.d1 )
        g3 = FEGrid( coord_max = ( 1., 1., 0. ),
                     shape = ( 2, 2 ),
                     fets_eval = self.fets_eval,
                     level = d3 )
        n_dofs = self.domain1.n_dofs
        #check the n_dofs of the domain after addition
        self.assertEqual( n_dofs, 54 )
        #check elem_dof_map of added subdomain
        elem_dof_map = d3.elem_dof_map

        edm = [36, 37, 42, 43, 44, 45, 38, 39,
               38, 39, 44, 45, 46, 47, 40, 41,
               42, 43, 48, 49, 50, 51, 44, 45,
               44, 45, 50, 51, 52, 53, 46, 47]

        for e_, e_ex_ in zip( elem_dof_map.flatten() , edm ):
            self.assertEqual( e_, e_ex_ )

class FEDomainAdaptiveRefinement( unittest.TestCase ):
    '''
    Test class for adaptive refinement:

    Using the FEDomain with single parent FERefinementGrid and one 
    child FERefinementGrid perform refinement of elements within 
    the two grids and test the activation and deactivation. This 
    should be done by verifying the n_dofs and elem_dof map obtained 
    from both grids. 
    '''
    pass

class FEDomainGeoMap( unittest.TestCase ):
    '''
    Test the retrieval of geometric information of FEDomain.
    '''
    def setUp( self ):
        '''
        Construct the FEDomain with one FERefinementGrids (2,2) 
        '''
        self.domain1 = FEDomain()
        self.fets_eval = FETS2D4Q( mats_eval = MATS2DElastic() )
        self.d1 = FERefinementGrid( name = 'd1', domain = self.domain1 )
        self.g1 = FEGrid( coord_max = ( 1., 1., 0. ),
                     shape = ( 2, 2 ),
                     fets_eval = self.fets_eval,
                     level = self.d1 )

    def test_elem_X_map( self ):
        '''Test the retrieval of geometric information of FEDomain.
        '''
        elem_X_map = self.g1.elem_X_map
        egm = [ 0., 0., 0.5, 0., 0.5, 0.5, 0., 0.5, 0., 0.5, 0.5, 0.5, 0.5, 1., 0., 1., \
               0.5, 0., 1., 0., 1., 0.5, 0.5, 0.5, 0.5, 0.5, 1., 0.5, 1., 1., 0.5, 1.]
        for e_, e_ex_ in zip( elem_X_map.flatten() , egm ):
            self.assertAlmostEqual( e_, e_ex_ )


class FEDomainSliceTest( unittest.TestCase ):
    '''
    Test the slicing of a simple FEGrid.
    '''
    def setUp( self ):
        '''
        Construct the FEDomain with one FERefinementGrids (2,2) 
        '''
        self.fets_eval = FETS3D8H()
        self.grid = FEGrid( coord_max = ( 1., 1., 1. ),
                     shape = ( 1, 1, 1 ),
                     fets_eval = self.fets_eval )

    def test_dof_slices( self ):
        '''Test the retrieval of dofs on a slice. 
        Check the dofs retrieved on a cube with a single element on every
        axis plane. 
        '''

        result = array( [[[ 0, 1, 2],
                       [12, 13, 14],
                       [ 6, 7, 8],
                       [18, 19, 20]],
                      [[ 0, 1, 2],
                       [12, 13, 14],
                       [ 3, 4, 5],
                       [15, 16, 17]],
                      [[ 0, 1, 2],
                       [ 6, 7, 8],
                       [ 3, 4, 5],
                       [ 9, 10, 11]]], dtype = int )

        fe_grid_slice = self.grid[:, :, 0, :, :, 0 ] # xy plane  1
        for dof_, dof_ex_ in zip( result[0].flatten() , fe_grid_slice.dofs.flatten() ):
            self.assertAlmostEqual( dof_, dof_ex_ )

        fe_grid_slice = self.grid[:, 0   , :, :, 0   , :] # xz plane  1
        for dof_, dof_ex_ in zip( result[1].flatten() , fe_grid_slice.dofs.flatten() ):
            self.assertAlmostEqual( dof_, dof_ex_ )

        fe_grid_slice = self.grid[0, :, :, 0   , :, :] # yz plane  1
        for dof_, dof_ex_ in zip( result[2].flatten() , fe_grid_slice.dofs.flatten() ):
            self.assertAlmostEqual( dof_, dof_ex_ )




from mayavi.core.ui.api import \
    MlabSceneModel
from traits.api import HasTraits, Instance, Button, \
    on_trait_change, Float, Property, cached_property, Bool, \
    Tuple
import bmcs_utils as bu
import numpy as np
from .yield_face3D import YieldConditionAbaqus, YieldConditionWillamWarnke, \
    YieldConditionDruckerPrager, YieldConditionVonMises, YieldConditionRankine


class YieldFaceViewer(bu.Model):

    min_sig = Float(-20.0, auto_set=False, enter_set=True, view_changed=True)
    max_sig = Float(5.0, auto_set=False, enter_set=True, view_changed=True)

    sig_i = Property(depends_on='min_sig, max_sig')

    @cached_property
    def _get_sig_i(self):
        n_sig = 100j
        sig_1, sig_2, sig_3 = np.mgrid[self.min_sig: self.max_sig: n_sig,
                                       self.min_sig: self.max_sig: n_sig,
                                       self.min_sig: self.max_sig: n_sig]
        return [sig_1, sig_2, sig_3]

    sig_ij = Property(depends_on='min_sig, max_sig')

    @cached_property
    def _get_sig_ij(self):
        sig_abcj = np.einsum('jabc->abcj', np.array(self.sig_i))
        DELTA = np.identity(3)
        sig_abcij = np.einsum('abcj,jl->abcjl', sig_abcj, DELTA)
        return sig_abcij

    yc_VM = Instance(YieldConditionVonMises, arg=(), kw={})
    yc_VM_on = Bool(True, view_changed=True)
    yc_DP = Instance(YieldConditionDruckerPrager, arg=(), kw={})
    yc_DP_on = Bool(False, view_changed=True)
    yc_R = Instance(YieldConditionRankine, arg=(), kw={})
    yc_R_on = Bool(False, view_changed=True)
    yc_abaqus = Instance(YieldConditionAbaqus, arg=(), kw={})
    yc_abaqus_on = Bool(False, view_changed=True)
    yc_WW = Instance(YieldConditionWillamWarnke, arg=(), kw={})
    yc_WW_on = Bool(False, view_changed=True)
    scene = Instance(MlabSceneModel)

    def _scene_default(self):
        return MlabSceneModel()

    bgcolor = Tuple(1.0, 1.0, 1.0)
    fgcolor = Tuple(0.0, 0.0, 0.0)

    mlab = Property(depends_on='input_change')
    '''Get the mlab handle'''

    def _get_mlab(self):
        return self.scene.mlab

    fig = Property()
    '''Figure for 3D visualization.
    '''
    @cached_property
    def _get_fig(self):
        fig = self.mlab.gcf()
        bgcolor = tuple(self.bgcolor)
        fgcolor = tuple(self.fgcolor)
        self.mlab.figure(fig, fgcolor=fgcolor, bgcolor=bgcolor)
        return fig

    button1 = Button('Redraw')

    @on_trait_change('button1,+view_changed')
    def redraw_scene(self):
        # Notice how each mlab call points explicitly to the figure it
        # applies to.
        self.fig
        self.mlab.clf()
        fig = self.mlab.gcf()
        fig.scene.disable_render = True

        if self.yc_abaqus_on:
            f = self.yc_abaqus.f(self.sig_ij)
            self.mlab.contour3d(
                self.sig_i[0], self.sig_i[1], self.sig_i[2], f,
                contours=[0.0], color=(1, 0, 0)
            )

        if self.yc_DP_on:
            f = self.yc_DP.f(self.sig_ij)
            self.mlab.contour3d(
                self.sig_i[0], self.sig_i[1], self.sig_i[2], f,
                contours=[0.0], color=(0, 1, 0)
            )

        if self.yc_WW_on:
            f = self.yc_WW.f(self.sig_ij)
            self.mlab.contour3d(
                self.sig_i[0], self.sig_i[1], self.sig_i[2], f,
                contours=[0.0], color=(0, 0, 1)
            )

        if self.yc_VM_on:
            f = self.yc_VM.f(self.sig_ij)
            f_pipe = self.mlab.contour3d(
                self.sig_i[0], self.sig_i[1], self.sig_i[2], f,
                contours=[0.0], color=(0, 0, 1)
            )

        if self.yc_R_on:
            f = self.yc_R.f(self.sig_ij)
            self.mlab.contour3d(
                self.sig_i[0], self.sig_i[1], self.sig_i[2], f,
                contours=[0.0], color=(0, 0.5, .5)
            )

        fig.scene.disable_render = False


def run_explorer(*args, **kw):
    m = YieldFaceViewer()
    m.redraw_scene()
    m.configure_traits(*args, **kw)



#-------------------------------------------------------------------------------
#
# Copyright (c) 2009, IMB, RWTH Aachen.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in simvisage/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.simvisage.com/licenses/BSD.txt
#

from ibvpy.fets import FETSEval

#-----------------------------------------------------------------------------------
# FETS3D - Base class for 3D elements 
#-----------------------------------------------------------------------------------

class FETS3D(FETSEval):
    '''Base class for 3D elements.
    '''






from traits.api import \
    Interface, Int, Array, Property


class IFEUniformDomain(Interface):
    '''
    Interface of the spatial domain.
    '''
    n_active_elems = Property(Int)

    elem_dof_map = Array

    elem_X_map = Array

    elem_x_map = Array

    def apply_on_ip_grid(self, fn, ip_mask):
        '''
        Apply the function fn over the first dimension of the array.
        @param fn: function to apply for each ip from ip_mask and each element. 
        @param ip_mask: specifies the local coordinates within the element.     
        '''



'''
Created on Mar 18, 2020

@author: rch
'''
import traits.api as tr
import bmcs_utils.api as bu

from .tline import TLine


class TLineMixIn(bu.Model):
    #=========================================================================
    # TIME LINE
    #=========================================================================
    tline = bu.Instance(TLine)
    r'''Time line defining the time range, discretization and state,  
    '''

    def _tline_default(self):
        return TLine(
            time_change_notifier=self.time_changed,
            time_range_change_notifier=self.time_range_changed
        )

    def time_changed(self, time):
        if not(self.ui is None):
            self.ui.time_changed(time)

    def time_range_changed(self, tmax):
        self.tline.max = tmax
        if not(self.ui is None):
            self.ui.time_range_changed(tmax)

    def set_tmax(self, time):
        self.time_range_changed(time)

    t = tr.DelegatesTo('tline', 'val')
    t_max = tr.DelegatesTo('tline', 'max')



#-------------------------------------------------------------------------------
#
# Copyright (c) 2009, IMB, RWTH Aachen.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in simvisage/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.simvisage.com/licenses/BSD.txt
#
# Thanks for using Simvisage open source!
#
# Created on Aug 19, 2009 by: rch
    
from .mats2D5_cmdm import \
    MATS2D5MicroplaneDamage, PhiFnStrainHardening

from ibvpy.tmodel.mats2D5.__test__ import \
    TestMATS2D5

class TestMATS2D5( TestMATS2D5 ):
    '''Test the elementary load cases of 2D5 material models.
    '''

    def _mats_eval_default(self):
        return MATS2D5MicroplaneDamage( E = 34000, nu = 0.25,
                                        phi_fn = PhiFnStrainHardening() )
        
    def test_symmetry_for_tension(self):
        '''
        Symmetry of the response ... inplane tensile loading.
        '''
        self.assert_2D_symmetry_clamped_cube( load_dirs = [0,1], 
                                              load = 0.0001 )        
        
    def test_symmetry_for_inplane_shear(self):
        '''
        Symmetry of the response ... inplane shear loading.
        '''
        self.assert_2D_symmetry_clamped_cube( load_dirs = [1,0], 
                                              load = 0.0002 )        

    def test_symmetry_for_bending(self):
        '''
        Symmetry of the response ... out of plane load.ing
        '''
        self.assert_2D_symmetry_clamped_cube( load_dirs = [2,2], 
                                              load = 0.001 )        

    def test_stress_value(self):
        '''Assert a correct value of resulting stress a material point.
        '''
        # @todo - response tracer should return 3D stress state - the current
        # value is an intermediate value that does not have transversal 
        # component equal to zero. This should be fixed by the 2D5 implementation
        # of CMDM.
        self.assert_stress_value( [ 3.0117268131744122, 0, 0, 0.35733911053546097 ],
                                    n_steps = 1, load = 0.0001 )
        
if __name__ == "__main__":
    import unittest
    import sys;sys.argv = ['', 'TestMATS2D5.test_symmetry_for_inplane_shear']
    unittest.main()


'''
Created on 30.10.2018

@author: abaktheer
'''
from ibvpy.tmodel.mats3D.mats3D_eval import \
    MATS3DEval
from traits.api import \
    List

import bmcs_utils.api as bu
import numpy as np
import traits.api as tr


class MATS3DDesmorat(MATS3DEval):
    '''Damage - plasticity model by Desmorat.
    '''

    #-------------------------------------------------------------------------
    # Material parameters
    #-------------------------------------------------------------------------

    E_1 = bu.Float(16.0e+3,
                   label="E_1",
                   desc="first Young's Modulus",
                   auto_set=False,
                   input=True)
    E_2 = bu.Float(19.0e+3,
                   label="E_2",
                   desc="second Young's Modulus",
                   auto_set=False,
                   input=True)

    nu = bu.Float(0.2,
                  label='nu',
                  desc="Poisson ratio",
                  auto_set=False,
                  input=True)

    gamma = bu.Float(110.0,
                  label="Gamma",
                  desc="kinematic hardening modulus",
                  MAT=True,
                  symbol=r'\gamma',
                  unit='MPa/mm',
                  enter_set=True,
                  auto_set=False)

    K = bu.Float(130.0,
              label="K",
              desc="isotropic hardening modulus",
              MAT=True,
              symbol='K',
              unit='MPa/mm',
              enter_set=True,
              auto_set=False)

    S = bu.Float(476.0e-6,
              label="S",
              desc="damage strength",
              MAT=True,
              symbol='S',
              unit='MPa/mm',
              enter_set=True,
              auto_set=False)

    tau_bar = bu.Float(6.0,
                    label="Tau_0 ",
                    desc="yield stress",
                    symbol=r'\bar{\tau}',
                    unit='MPa',
                    MAT=True,
                    enter_set=True,
                    auto_set=False)

    #=========================================================================
    # Configurational parameters
    #=========================================================================
    U_var_shape = (6,)
    '''Shape of the primary variable required by the TStepState.
    '''

    state_var_shapes = {'sigma_ab': (3, 3),
                        'sigma_pi_ab': (3, 3),
                        'eps_pi_ab': (3, 3),
                        'alpha_ab': (3, 3),
                        'z_a': (),
                        'omega_a': ()}
    r'''
    Shapes of the state variables
    to be stored in the global array at the level 
    of the domain.
    '''

    node_name = 'Desmorat model'

    tree_node_list = List([])

    def _get_lame_1_params(self):
        la = self.E_1 * self.nu / ((1. + self.nu) * (1. - 2. * self.nu))
        # second Lame parameter (shear modulus)
        mu = self.E_1 / (2. + 2. * self.nu)
        return la, mu

    D_1_abef = tr.Property(tr.Array, depends_on='+input')

    @tr.cached_property
    def _get_D_1_abef(self):
        la, mu = self._get_lame_1_params()
        delta = np.identity(3)
        D_1_abef = (np.einsum(',ij,kl->ijkl', la, delta, delta) +
                    np.einsum(',ik,jl->ijkl', mu, delta, delta) +
                    np.einsum(',il,jk->ijkl', mu, delta, delta))

        return D_1_abef

    def _get_lame_2_params(self):
        la = self.E_2 * self.nu / ((1. + self.nu) * (1. - 2. * self.nu))
        # second Lame parameter (shear modulus)
        mu = self.E_2 / (2. + 2. * self.nu)
        return la, mu

    D_2_abef = tr.Property(tr.Array, depends_on='+input')

    @tr.cached_property
    def _get_D_2_abef(self):
        la = self._get_lame_2_params()[0]
        mu = self._get_lame_2_params()[1]
        delta = np.identity(3)
        D_2_abef = (np.einsum(',ij,kl->ijkl', la, delta, delta) +
                    np.einsum(',ik,jl->ijkl', mu, delta, delta) +
                    np.einsum(',il,jk->ijkl', mu, delta, delta))

        return D_2_abef

    def get_corr_pred(self, eps_ab, tn1,
                      sigma_ab, sigma_pi_ab, eps_pi_ab,
                      alpha_ab, z_a, omega_a):
        r'''
        Corrector predictor computation.
        '''
        D_1_abef = self.D_1_abef
        D_2_abef = self.D_2_abef
        sigma_pi_ab_trial = np.einsum(
            '...ijkl,...kl->...ij',
            D_2_abef, eps_ab - eps_pi_ab
        )
        a = sigma_pi_ab_trial - self.gamma * alpha_ab
        norm_a = np.sqrt(np.einsum(
            '...ij,...ij',
            a, a)
        )
        f = norm_a - self.tau_bar - self.K * z_a
        # identify the inelastic material points to perform return mapping
        I = np.where(f > 1e-6)
        delta_pi_I = (
            f[I] /
            (self.E_2 + (self.K + self.gamma) * (1. - omega_a[I]))
        )
        b_I = norm_a[I]
        return_ab_I = np.einsum(
            '...ij,...->...ij',
            a[I], delta_pi_I / b_I
        )
        eps_pi_ab[I] += return_ab_I
        eps_diff_ab_I = eps_ab[I] - eps_pi_ab[I]
        Y_a_I = 0.5 * (
            np.einsum(
                '...ij,...ijkl,...kl',
                eps_ab[I], D_1_abef, eps_ab[I]
            )
            +
            np.einsum(
                '...ij,...ijkl,...kl',
                eps_diff_ab_I, D_2_abef, eps_diff_ab_I
            )
        )
        omega_a[I] += (Y_a_I / self.S) * delta_pi_I
        omega_a[I][np.where(omega_a[I] >= 0.99)] = 0.99
        alpha_ab[I] += np.einsum(
            '...ij,...->...ij',
            return_ab_I, (1.0 - omega_a[I])
        )
        z_a[I] += delta_pi_I * (1.0 - omega_a[I])

        # evaluate the material stress and stiffness tensors
        phi_n = 1.0 - omega_a
        # this is a side effect - recording a returned value
        # simultaneously as a state variable - not ideal! dangerous.
        sigma_ab[...] = (
            np.einsum(
                '...,...ijkl,...kl->...ij',
                phi_n, D_1_abef, eps_ab
            ) +
            np.einsum(
                '...,...ijkl,...kl->...ij',
                phi_n, D_2_abef, eps_ab - eps_pi_ab
            )
        )
        # secant stiffness matrix
        D_abef = np.einsum(
            '...,...ijkl->...ijkl',
            phi_n, D_1_abef + D_2_abef
        )
        return sigma_ab, D_abef

    ipw_view = bu.View(
            bu.Item('E_1'),
            bu.Item('E_2'),
            bu.Item('nu'),
            bu.Item('gamma'),
            bu.Item('K'),
            bu.Item("S"),
            bu.Item("tau_bar"),
    )



#-------------------------------------------------------------------------
#
# Copyright (c) 2009, IMB, RWTH Aachen.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in simvisage/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.simvisage.com/licenses/BSD.txt
#
# Thanks for using Simvisage open source!
#
# Created on May 26, 2009 by: rchx

from .i_bcond import \
    IBCond
from ibvpy.mesh.fe_grid_idx_slice import FEGridIdxSlice
# from ibvpy.view.plot3d.mayavi_util.pipelines import \
#     MVPointLabels
from ibvpy.tfunction import TimeFunction, TFMonotonic
from ibvpy.mathkit.mfn import MFnLineArray
from numpy import \
    ix_, dot, repeat, zeros
from scipy.linalg import \
    det, norm
from traits.api import Float, \
    Instance, Int, Trait, Str, Enum, \
    List, cached_property, \
    Button, \
    provides, Property
from ibvpy.view.plot2d import Vis2D, Viz2DTimeFunction
from ibvpy.view.ui import BMCSTreeNode

import numpy as np

from .bc_dof import BCDof


@provides(IBCond)
class BCSlice(BMCSTreeNode, Vis2D):
    '''
    Implements the IBC functionality for a constrained dof.
    '''
    name = Str('<unnamed>')
    node_name = Property

    def _get_node_name(self):
        s = '%s:%s=%s' % (self.var, self.slice, self.value)
        return s

    var = Enum('u', 'f', 'eps', 'sig')

    # slice = Instance(FEGridIdxSlice)
    slice = Trait()

    link_slice = Instance(FEGridIdxSlice)

    bcdof_list = List(BCDof)

    def reset(self):
        self.bcdof_list = []
    # List of dofs that determine the value of the current dof
    #
    # If this list is empty, then the current dof is
    # prescribed. Otherwise, the dof value is given by the
    # linear combination of DOFs in the list (see the example below)
    #
    # link_dofs = List( Int )

    # Coefficients of the linear combination of DOFs specified in the
    # above list.
    #
    link_coeffs = List(Float)

    dims = List(Int)

    _link_dims = List(Int)
    link_dims = Property(List(Int))

    def _get_link_dims(self):
        if len(self._link_dims) == 0:
            return self.dims
        else:
            return self._link_dims

    def _set_link_dims(self, link_dims):
        self._link_dims = link_dims

    value = Float

    integ_domain = Enum(['global', 'local'])

    time_function = Instance(TimeFunction, ())

    def _time_function_default(self):
        return TFMonotonic()

    space_function = Instance(MFnLineArray, ())

    def _space_function_default(self):
        return MFnLineArray(xdata=[0, 1], ydata=[1, 1], extrapolate='diff')

    def is_essential(self):
        return self.var == 'u'

    def is_linked(self):
        return self.link_dofs != []

    def is_constrained(self):
        '''
        Return true if a DOF is either explicitly prescribed or it depends on other DOFS.
        '''
        return self.is_essential() or self.is_linked()

    def is_natural(self):
        return self.var == 'f' or self.var == 'eps' or self.var == 'sig'

    def setup(self, sctx):
        '''
        Locate the spatial context.f
        '''
        if self.link_slice == None:
            for el, el_dofs, el_dofs_X in zip(self.slice.elems,
                                              self.slice.dofs,
                                              self.slice.dof_X):
                # print 'el_dofs', el_dofs
                for node_dofs, dof_X in zip(el_dofs, el_dofs_X):
                    # print 'node_dofs ', node_dofs
                    for dof in node_dofs[self.dims]:
                        self.bcdof_list.append(BCDof(var=self.var,
                                                     dof=dof,
                                                     value=self.value,
                                                     link_coeffs=self.link_coeffs,
                                                     time_function=self.time_function))
        else:
            # apply the linked slice
            n_link_nodes = len(self.link_slice.dof_nodes.flatten())
            link_dofs = self.link_dofs
            if n_link_nodes == 1:
                #
                link_dof = self.link_slice.dofs.flatten()[0]
                link_coeffs = self.link_coeffs
                for el, el_dofs, el_dofs_X in \
                        zip(self.slice.elems, self.slice.dofs, self.slice.dof_X):
                    for node_dofs, dof_X in zip(el_dofs, el_dofs_X):
                        for dof, link_dof, link_coeff in zip(node_dofs[self.dims], link_dofs, link_coeffs):
                            self.bcdof_list.append(BCDof(var=self.var,
                                                         dof=dof,
                                                         link_dofs=[link_dof],
                                                         value=self.value,
                                                         link_coeffs=[
                                                             link_coeff],
                                                         time_function=self.time_function))
            else:
                for el, el_dofs, el_dofs_X, el_link, el_link_dofs, el_link_dofs_X in \
                    zip(self.slice.elems, self.slice.dofs, self.slice.dof_X,
                        self.link_slice.elems, self.link_slice.dofs, self.link_slice.dof_X):
                    # the link slice is compatible with the bc slice
                    #print('el', el, el_dofs, el_link, el_link_dofs)

                    for node_dofs, dof_X, node_link_dofs, link_dof_X in \
                            zip(el_dofs, el_dofs_X, el_link_dofs, el_link_dofs_X):

                        #print('node', node_dofs, node_link_dofs)
                        #print('node[dims]', node_dofs[self.dims],
                              # node_link_dofs[self.link_dims])
                        for dof, link_dof, link_coeff in zip(node_dofs[self.dims],
                                                             node_link_dofs[
                                                                 self.link_dims],
                                                             self.link_coeffs):
                            #print('dof, link, coeff', dof, link_dof, link_coeff)
                            self.bcdof_list.append(BCDof(var=self.var,
                                                         dof=dof,
                                                         link_dofs=[link_dof],
                                                         value=self.value,
                                                         link_coeffs=[
                                                             link_coeff],
                                                         time_function=self.time_function))

    def register(self, K):
        '''Register the boundary condition in the equation system.
        '''
        for bcond in self.bcdof_list:
            bcond.register(K)

    def apply_essential(self, K):

        for bcond in self.bcdof_list:
            bcond.apply_essential(K)

    def apply(self, step_flag, sctx, K, R, t_n, t_n1):

        if self.is_essential():
            for bcond in self.bcdof_list:
                bcond.apply(step_flag, sctx, K, R, t_n, t_n1)
        else:
            self.apply_natural(step_flag, sctx, K, R, t_n, t_n1)

    def apply_natural(self, step_flag, sctx, K, R, t_n, t_n1):

        fets_eval = self.slice.fe_grid.fets_eval

        e_idx, n_idx = self.slice.dof_grid_slice.idx_tuple

        r_arr, w_arr, ix = fets_eval.get_sliced_ip_scheme(n_idx)

        slice_geo_X = self.slice.fe_grid.elem_X_map[self.slice.elems]
        slice_dofs = self.slice.fe_grid.elem_dof_map[self.slice.elems]

        p_value = self.value * float(self.time_function(t_n1))

        p_vct = zeros((fets_eval.n_nodal_dofs,), dtype='float64')
        for d in self.dims:
            p_vct[d] = p_value

        for el, el_dofs, el_geo_X in zip(self.slice.elems, slice_dofs, slice_geo_X):
            f_vct = zeros((fets_eval.n_e_dofs,), dtype='float64')
            for r_pnt, w in zip(r_arr, w_arr):
                if len(ix) > 0:
                    J_mtx = fets_eval.get_J_mtx(r_pnt, el_geo_X)

                    if self.integ_domain == 'global':

                        # Integration projected onto the global coordinates
                        # axes - use submatrix of the J_mtx
                        #
                        J_det = det(J_mtx[ix_(ix, ix)])

                    elif self.integ_domain == 'local':

                        # Integration over the parametric coordinates on the
                        # surface
                        #
                        J_det = 1.0
                        for i in ix:
                            J_det *= norm(J_mtx[:, i])
                else:
                    J_det = 1.0

                sctx.r_pnt = r_pnt
                sctx.X = el_geo_X
                X_pnt = fets_eval.get_X_pnt(sctx)
                space_factor = self.space_function(X_pnt)

                # add to the element vector
                N_mtx = fets_eval.get_N_mtx(r_pnt)
                f_vct += dot(N_mtx.T, p_vct * space_factor) * w * J_det

            R[el_dofs] += f_vct

    #-------------------------------------------------------------------------
    # Ccnstrained DOFs
    #-------------------------------------------------------------------------

    dofs = Property

    def _get_dofs(self):
        return np.unique(self.slice.dofs[..., self.dims].flatten())

    dof_X = Property

    def _get_dof_X(self):
        return self.slice.dof_X

    n_dof_nodes = Property

    def _get_n_dof_nodes(self):
        sliceshape = self.dofs.shape
        return sliceshape[0] * sliceshape[1]

    # # register the pipelines for plotting labels and geometry
    # #
    # mvp_dofs = Trait(MVPointLabels)
    #
    # def _mvp_dofs_default(self):
    #     return MVPointLabels(name='Boundary condition',
    #                          points=self._get_mvpoints,
    #                          vectors=self._get_labels,
    #                          color=(0.0, 0.0, 0.882353))

    def _get_mvpoints(self):
        # blow up
        if self.dof_X.shape[2] == 2:
            dof_X = np.hstack([self.dof_X.reshape(self.n_dof_nodes, 2),
                               np.zeros((self.n_dof_nodes, 1), dtype='f')])
        elif self.dof_X.shape[2] == 1:
            dof_X = np.hstack([self.dof_X.reshape(self.n_dof_nodes, 1),
                               np.zeros((self.n_dof_nodes, 2), dtype='f')])
        else:
            dof_X = self.dof_X.reshape(self.n_dof_nodes, 3)
        return dof_X

    def _get_labels(self):
        # blow up
        n_points = self.n_dof_nodes
        dofs = repeat(-1., n_points * 3).reshape(n_points, 3)
        dofs[:, tuple(self.dims)] = self.dofs
        return dofs

    #-------------------------------------------------------------------------
    # Link DOFs
    #-------------------------------------------------------------------------
    link_dofs = Property(List)

    def _get_link_dofs(self):
        if self.link_slice != None:
            return np.unique(self.link_slice.dofs[..., self.link_dims].flatten())
        else:
            return []

    link_dof_X = Property

    def _get_link_dof_X(self):
        return self.link_slice.dof_X

    n_link_dof_nodes = Property

    def _get_n_link_dof_nodes(self):
        sliceshape = self.link_dofs.shape
        return sliceshape[0] * sliceshape[1]

    # # register the pipelines for plotting labels and geometry
    # #
    # mvp_link_dofs = Trait(MVPointLabels)
    #
    # def _mvp_link_dofs_default(self):
    #     return MVPointLabels(name='Link boundary condition',
    #                          points=self._get_link_mvpoints,
    #                          vectors=self._get_link_labels,
    #                          color=(0.0, 0.882353, 0.0))

    def _get_link_mvpoints(self):
        # blow up
        # blow up
        if self.link_slice == None:
            return np.zeros((0, 3), dtype='f')
        if self.dof_X.shape[2] == 2:
            dof_X = np.hstack([self.link_dof_X.reshape(self.n_dof_nodes, 2),
                               np.zeros((self.n_dof_nodes, 1), dtype='f')])
        elif self.dof_X.shape[2] == 1:
            dof_X = np.hstack([self.link_dof_X.reshape(self.n_dof_nodes, 1),
                               np.zeros((self.n_dof_nodes, 2), dtype='f')])
        else:
            dof_X = self.link_dof_X.reshape(self.n_dof_nodes, 3)
        return dof_X

    def _get_link_labels(self):
        # blow up
        if self.link_slice == None:
            return np.zeros((0, 3), dtype='f')
        n_points = self.n_link_dof_nodes
        dofs = repeat(-1., n_points * 3).reshape(n_points, 3)
        dofs[:, tuple(self.dims)] = self.link_dofs
        return dofs

    redraw_button = Button('Redraw')

    def _redraw_button_fired(self):
        self.mvp_dofs.redraw(label_mode='label_vectors')
        self.mvp_link_dofs.redraw(label_mode='label_vectors')

    viz2d_classes = {'time function': Viz2DTimeFunction}



from traits.api import \
    HasTraits, provides, \
    Instance, Int, Str, \
    on_trait_change, Property, \
    self, TraitError
from .fe_domain import FEDomain
from .i_fe_subdomain import IFESubDomain


_subdomain_counter = 0


@provides(IFESubDomain)
class FESubDomain(HasTraits):

    # If the constructor specifies the name, then use it
    # otherwise generate the name based on the domain enumeration.
    _name = Str('_no_name_')
    _tree_label = Str('subdomain')

    tstepper = Property

    def _get_tstepper(self):
        return self.domain.dots.tstepper

    # specification of the container domain
    # managing all the sub domains
    _domain = Instance(FEDomain)
    domain = Property

    def _set_domain(self, value):
        'reset the domain of this domain'
        if self._domain:
            # unregister in the old domain
            raise NotImplementedError(
                'FESubDomain cannot be relinked to another FEDomain')

        self._domain = value
        # register in the domain as a sub domain
        self._domain.subdomains.append(self)
        self._domain._append_in_series(self)

    def _get_domain(self):
        return self._domain

    def validate(self):
        pass

    name = Property(Str)

    def _get_name(self):
        global _subdomain_counter
        if self._name == '_no_name_':
            self._name = self._tree_label + ' ' + str(_subdomain_counter)
            _subdomain_counter += 1
        return self._name

    def _set_name(self, value):
        self._name = value

    # local dof enumeration
    n_dofs = Int(4, domain_changed=True)

    # dof offset within the global enumeration
    dof_offset = Property(Int, depends_on='previous_domain.dof_offset')
    # cached_property

    def _get_dof_offset(self):
        if self.previous_domain:
            return self.previous_domain.dof_offset + self.previous_domain.n_dofs
        else:
            return 0

    def __repr__(self):
        if self.previous_domain:
            return self.name + ' <- ' + self.previous_domain.name
        else:
            return self.name

    # get the slice for DOFs within global vectors
    sub_slice = Property

    def _get_sub_slice(self):
        return slice(self.dof_offset, self.dof_offset + self.n_dofs)

    # dependency link for sequential enumeration
    previous_domain = Instance(IFESubDomain, domain_changed=True)

    @on_trait_change('previous_domain')
    def _validate_previous_domain(self):
        if self.previous_domain == self:
            raise TraitError('cyclic reference for ' + self.name)

    # dependency link for sequential enumeration
    next_domain = Instance(IFESubDomain, domain_changed=True)

    @on_trait_change('next_domain')
    def _validate_next_domain(self):
        if self.next_domain == self:
            raise TraitError('cyclic reference for ' + self.name)




'''
Created on Mar 4, 2017

@author: rch
'''

import numpy as np
from ibvpy.view.plot2d import Vis2D, Viz2D
from ibvpy.view.ui.bmcs_tree_node import BMCSLeafNode


class TP(Viz2D):

    def plot(self, ax, vot):
        t, x, y = self.vis2d.get_sim_results(vot)
        ax.plot(x, y)


class ResponseTracer(BMCSLeafNode, Vis2D):

    node_name = 'response tracer'

    def get_sim_results(self, vot):
        t = np.linspace(0, 1, 100)
        x = t
        y = x**2 - vot
        return t, x, y

    viz2d_classes = {
        'time_profile': TP
    }




from math import fabs

import numpy as np
from traits.api import \
    HasTraits, Int, Array, Property, cached_property, List, Trait, Dict, \
    Any, Bool, Float

from .coo_mtx import COOSparseMtx
from .sys_mtx_array import SysMtxArray

class Constraint(HasTraits):

    u_a = Property(Float)

    def _set_u_a(self, value):
        self._u_a = value * self._freeze_coeff

    def _get_u_a(self):
        return self._u_a

    # Freeze the constraint by setting the coefficient to zero
    # applied for overruled constraints
    #
    _freeze_coeff = 1.0

    def freeze(self):
        self._freeze_coeff = 0.0

    a = Int(-1)
    alpha = Array(float)
    ix_a = Array(int)

    def __str__(self):
        return 'a = %d, u_a = %g, alpha = %s, ix_a = %s' % \
            (self.a, self.u_a, str(self.alpha), str(self.ix_a))


class SysMtxAssembly(HasTraits):
    '''Class intermediating between the time stepper and the linear algebra solver.

    Handling of constraints.

    Constraints are added by the method

        register_constraint( self, a, u_a, alpha, ix_a )

    They are registered in the trait list called

        constraints = List

    The inclusion of constraints is performed on the mapped matrices stored
    in the sys_mtx_arrays. It is possible that contributions to the position in
    the system matrix are stored in several sys_mtx_arrays. Therefore, the
    constraint must be included in all matrix arrays currently included.

    Such an operation requires a search through the matrices, which is realized
    using the

        numpy.where

    method. In order to avoid repeated calls of where the access to the
    affected positions in the sys_mtx_array can be cached. The state management
    recognizes the attributes

        value_changed = Bool
        structure_changed = Bool

    Thus, an adaptive strategy or spatial domain must indicate the change in
    the structure explicitly by setting

        sys_assembly = True

    The sys_assembly maintains a property attribute

        cached_constraints = Property()

    containing the result of the where operations delivering the addresses
    of the format

        [el, row]
    '''
    # list of matrix arrays
    #
    sys_mtx_arrays = List

    # list of matrix arrays
    #
    link_matrices = List

    # list of applied constraints
    #
    constraints = List

    # Dictionary for fast access
    #
    _c = Dict

    # number of degrees of freedom
    #
    n_dofs = Property(Int)

    def _get_n_dofs(self):
        n_dofs = 0
        # must iterate over all arrays to get the maximum number of DOFs
        # the enumeration is done within the spatial domain object.
        #
        for mtx_array in self.sys_mtx_arrays:
            n_dofs = max(n_dofs, mtx_array.n_dofs)
        return n_dofs

    def add_mtx_array(self, mtx_arr, dof_map_arr):
        '''Add an array of matrices with the dof map
        '''
        sys_mtx_arr = SysMtxArray(dof_map_arr=dof_map_arr,
                                  mtx_arr=mtx_arr)
        self.sys_mtx_arrays.append(sys_mtx_arr)

    def add_mtx(self, mtx, dof_map=None):
        '''Add a single matrix with the dof map
        '''
        if not type(dof_map) is np.ndarray:
            dof_map = np.arange(mtx.shape[0])

        sys_mtx_array = SysMtxArray(dof_map_arr=dof_map[None, ...],
                                    mtx_arr=mtx[None, ...]
                                    )
        self.sys_mtx_arrays.append(sys_mtx_array)
        return sys_mtx_array

    def add_link_mtx(self, mtx, dof_map=None):
        '''Add a single matrix with the dof map
        '''
        if not type(dof_map) is np.ndarray:
            dof_map = np.arange(mtx.shape[0])

        link_mtx = SysMtxArray(dof_map_arr=dof_map[None, ...],
                               mtx_arr=mtx[None, ...]
                               )
        self.link_matrices.append(link_mtx)

    debug = Bool(False)

    def register_constraint(self, a, u_a=0, alpha=[], ix_a=[]):

        constraint = self._c.get(a, None)

        if constraint == None:

            # constraint does not exist yet, add it
            #
            constraint = Constraint(a=a, u_a=u_a, alpha=alpha, ix_a=ix_a)
            self.constraints.append(constraint)
            self._c[a] = constraint

            if self.debug:
                print('new constraint: a:', a, 'u_a',
                      u_a, 'alpha', alpha, 'ix_a', ix_a)

        else:
            # duplicate specification, if it is identical with the
            # previous one simply ignore it, otherwise throw and exception.
            #
            # Handle the "corner" and "edge" constraints. The rule is, the
            # zero-constraint overrules the non-zero or a link constraint.
            if constraint.u_a == 0 and constraint.alpha.shape[0] == 0 and \
                    constraint.ix_a.shape[0] == 0:
                # ignore the non-zero constraint
                # the zero-constraint prevails
                #
                constraint.freeze()
                if self.debug:
                    print('frozen constraint:', constraint)
            elif u_a == 0 and len(alpha) == 0 and len(ix_a) == 0:
                # set the existing constraint to zero-constraint
                # ignore the non-zero value and the coefficients
                constraint.freeze()
                constraint.u_a = u_a
                constraint.alpha = alpha
                constraint.ix_a = ix_a
                if self.debug:
                    print('frozen constraint:', constraint)
            elif not constraint.u_a == u_a or \
                    not np.allclose(constraint.alpha, alpha, rtol=1e-4) or \
                    not np.array_equal(constraint.ix_a, ix_a):
                raise ValueError('contradicting constraint definition:\n'
                                 'a = %d, u = %f, alpha = %s, ix_a = %s\n'
                                 'previous constraint:\n%s' % (
                                     a, u_a, alpha, ix_a, constraint))

        return constraint

    sorted_constraints = Property(depends_on='constraints[]')

    @cached_property
    def _get_sorted_constraints(self):

        sorted_c = self.constraints
        for i in range(1000):
            sorted_c, swapped_values = self._get_simply_sorted_c(sorted_c)
            if swapped_values == False:
                return sorted_c
        # exception - the cyclic specification of constraints
        raise ValueError('maximum number of reorderings (1000) reached\n'
                         'this is probably due to cyclic constraints specification')

    def _get_simply_sorted_c(self, constraints):
        # first test if one of the indices in ix_a is already in _c
        sorted_a = []
        sorted_c = []
        swapped_values = False
        for constraint in constraints:
            # suggest the last position in the traversed elements
            pos = len(sorted_a)
            for ix in constraint.ix_a:
                try:
                    found_pos = sorted_a.index(ix)
                    if found_pos < pos:
                        # The found position is below the last one
                        # swap required. Insert the current constraint
                        # before the found one
                        pos = found_pos
                        swapped_values = True
                except ValueError:
                    pass
            sorted_a.insert(pos, constraint.a)
            sorted_c.insert(pos, constraint)

        return sorted_c, swapped_values

    cached_addresses = Property(depends_on='constraints[]')

    @cached_property
    def _get_cached_addresses(self):
        cached_addresses = []
        for constraint in self.sorted_constraints:
            ix_maps = self._get_ix_maps(constraint)
            cached_addresses.append(ix_maps)
        return cached_addresses

    def solve(self, rhs=None, check_pos_def=False):
        '''Solve the system of equations using a specified
        type of matrix format
        '''

        if self.debug:
            print('SysMtxAssembly:', id(self))

        if rhs is None and self._rhs is None:
            raise ValueError('No right hand side available')

        if not rhs is None:
            self.apply_constraints(rhs)

        mtx = COOSparseMtx(assemb=self)
        return mtx.solve(self._rhs, check_pos_def)

    def reset(self):
        self.sys_mtx_arrays = []
        self.constraints = []
        self.link_matrices = []
        self._c = {}
        self.rhs = None

    def reset_mtx(self):
        self.sys_mtx_arrays = []
        self.rhs = None

    _rhs = Any

    def print_constraints(self):
        # apply the constraints
        for constraint in self.constraints:
            print(constraint)

    def apply_constraints(self, rhs):
        # apply the constraints
        for constraint, ix_maps in zip(self.sorted_constraints,
                                       self.cached_addresses):
            if self.debug:
                print('applying constraint', constraint)

            self._apply_constraint(rhs, constraint, ix_maps)

        self._rhs = rhs  # rhs that should be used for solving the system

    def get_sys_mtx_arrays(self):
        '''Return the complete list of matrix arrays including constraints.
        '''
        return self.sys_mtx_arrays + self.link_matrices

    coo_matrix = Property()
    def _get_coo_matrix(self):
        return COOSparseMtx(assemb=self)

    def toarray(self):
        '''Create dense matrix and print it
        '''
        return self.coo_matrix.toarray()

    def _get_ix_maps(self, constraint):
        '''
        Add the constraint associated with the dof number `a'
        '''
        c = constraint

        dof_ix_array = self.get_dof_ix_array(c.a)

        a, u_a, alpha, ix_a = (c.a, c.u_a, c.alpha, c.ix_a)

        # link

        if type(alpha) == list:
            alpha = np.array(alpha, dtype=float)
        if type(ix_a) == list:
            ix_a = np.array(ix_a, dtype=int)

        if alpha.shape[0] == 0:

            # no links - manipulates only a single equation
            #
            return (alpha, None, dof_ix_array, None, None, None, None, None, None, None)

        # find out which non-zero entries has the affected dofs
        ix_K, K_n_a = self._get_col_subvector(dof_ix_array)

        ix_orig_layout = np.hstack([ix_K, ix_a])

#            K_n_a2 = zeros( ix_layout.shape[0], dtype = float )
#            K_n_a2[:ix_K.shape[0]] = K_n_a
#            alpha2 = zeros( ix_layout.shape[0], dtype = float )
#            alpha2[ix_K.shape[0]:] = alpha

        ix_mask = ix_orig_layout != a  # deactivate the constrained dof

        prev_same_i_list = []
        alpha_same_i_list = []
        for i, ix in enumerate(ix_K):
            prev_same_i = np.where(ix_K == ix)[0][0]
            prev_same_i_list.append(prev_same_i)
            if prev_same_i < i:
                #                    K_n_a2[prev_same_i] += K_n_a2[i]
                ix_mask[i] = False  # deactivate the added value

            if ix in ix_a:
                alpha_same_i = np.where(ix_a == ix)[0][0]
                alpha_same_i_list.append(alpha_same_i)
                alpha_same_i += ix_K.shape[0]
#                    alpha2[i] = alpha2[ alpha_same_i ]
                ix_mask[alpha_same_i] = False
            else:
                alpha_same_i_list.append(0)

#            K_n_a     = compress( ix_mask, K_n_a2 )
        ix_layout = np.compress(ix_mask, ix_orig_layout)
#            alpha     = compress( ix_mask, alpha2 )

        # Get the size of the link matrix
        link_mtx_sz = ix_layout.shape[0] + 1

        #
        # Add the a-th column to the right hand side
        # Note, K_n_a is not the full column, it contains only the
        # rows of the dof involved in the linear combination (indices)

#            P_a = rhs[a]
#            rhs[a] = 0

#            self._add_col_to_vector( dof_ix_array, rhs, factor = - u_a )

#            K_a_a = self._get_diag_elem( dof_ix_array )

        # After remembering the submatrices in K_aa, and K_na
        # the constrained row and column can be zeroed
        #
#            self._zero_rc( dof_ix_array )

        # Redistribute the load applied to the constrained dof
        #
#            rhs[ ix_layout ] += alpha.transpose() * P_a

        # fill the dof map of the additional link matrix
        #
        link_dof_map = np.zeros((link_mtx_sz), dtype=int)
        link_dof_map[0] = a
        link_dof_map[1:] = ix_layout

        # fill the link matrix itself
        #
        # @todo: split -
        link_mtx = np.zeros((link_mtx_sz, link_mtx_sz), dtype=float)
#        link_mtx[0,0] = - K_a_a
#        link_mtx[0,1:] = K_a_a * alpha.transpose()
#        link_mtx[1:,0] = link_mtx[0,1:].transpose()
#        K_n_a_alpha = dot( K_n_a[:,None], alpha[None,:] )
#        link_mtx[1:,1:] = K_n_a_alpha + K_n_a_alpha.transpose()

        # Add the link matrix
        #
        self.add_link_mtx(mtx=link_mtx, dof_map=link_dof_map)

        return (alpha, ix_a, dof_ix_array, link_dof_map,
                ix_orig_layout, ix_layout, ix_mask, link_mtx,
                np.array(prev_same_i_list, dtype='int_'), np.array(alpha_same_i_list, dtype='int_'))

    def _apply_constraint(self, rhs, constraint, ix_map):
        '''
        Add the constraint associated with the dof number `a'
        '''
        c = constraint

        alpha, ix_a, dof_ix_array, link_dof_map, ix_orig_layout, ix_layout, ix_mask, link_mtx, prev_same_i_array, alpha_same_i_array = ix_map

        a, u_a, alpha, ix_a = (c.a, c.u_a, c.alpha, c.ix_a)

        # link

        if alpha.shape[0] == 0:

            # constraint affect only a single equation
            # simplified handling.
            #
            # Add the a-th column to the right hand side
            # Note, K_n_a is not the full column, it contains only the
            # rows of the dof involved in the linear combination (indices)

            P_a = rhs[a]
            rhs[a] = 0

            self._add_col_to_vector(dof_ix_array, rhs, factor=-u_a)

            K_aa = self._get_diag_elem(dof_ix_array)

            if fabs(K_aa) < (1.0e-5):
                K_aa = 1.
                rhs[a] = -u_a

            # After remembering the submatrices in K_aa, and K_na
            # the constrained row and column can be zeroed
            #
            self._zero_rc(dof_ix_array)

            self._add_diag_elem(dof_ix_array, -K_aa)

        else:

            # find out which non-zero entries has the affected dofs
            ix_K, K_n_a = self._get_col_subvector(dof_ix_array)

            K_n_a2 = np.zeros(ix_orig_layout.shape[0], dtype=float)
            K_n_a2[:ix_K.shape[0]] = K_n_a

            alpha2 = np.zeros(ix_orig_layout.shape[0], dtype=float)
            alpha2[ix_K.shape[0]:] = alpha

            # ix_mask = ix_layout != a # deactivate the constrained dof

            for i, ix in enumerate(ix_K):
                #prev_same_i = where( ix_K == ix )[0][0]
                prev_same_i = prev_same_i_array[i]
                if prev_same_i < i:
                    K_n_a2[prev_same_i] += K_n_a2[i]

                if ix in ix_a:
                    #alpha_same_i = where( ix_a == ix )[0][0]
                    alpha_same_i = alpha_same_i_array[i]
                    alpha_same_i += ix_K.shape[0]
                    alpha2[i] = alpha2[alpha_same_i]

            K_n_a = np.compress(ix_mask, K_n_a2)
            alpha = np.compress(ix_mask, alpha2)

            # Get the size of the link matrix
            link_mtx_sz = alpha.shape[0] + 1

            #
            # Add the a-th column to the right hand side
            # Note, K_n_a is not the full column, it contains only the
            # rows of the dof involved in the linear combination (indices)

            P_a = rhs[a]
            rhs[a] = 0

            self._add_col_to_vector(dof_ix_array, rhs, factor=-u_a)

            K_a_a = self._get_diag_elem(dof_ix_array)

            # This has not been thoroughly tested
            # If a DOF is in the air, it can be linked to an existing
            # dof via kinematic constraint. This is used upon deactivation
            # of elements.
            #
            if fabs(K_a_a) < (1.0e-5):
                K_a_a = 1.

            # After remembering the submatrices in K_aa, and K_na
            # the constrained row and column can be zeroed
            #
            self._zero_rc(dof_ix_array)

            # Redistribute the load applied to the constrained dof
            #
            rhs[ix_layout] += alpha.transpose() * P_a

            # fill the link matrix itself
            #
            #link_mtx = zeros( (link_mtx_sz, link_mtx_sz), dtype = float )
            link_mtx[0, 0] = -K_a_a
            link_mtx[0, 1:] = K_a_a * alpha.transpose()
            link_mtx[1:, 0] = link_mtx[0, 1:].transpose()
            K_n_a_alpha = np.dot(K_n_a[:, None], alpha[None, :])
            link_mtx[1:, 1:] = K_n_a_alpha + K_n_a_alpha.transpose()

            # Add the link matrix
            #
            # This has been moved to the cached part _get_ix_maps
            # - the link matrices are only modified in-place.
            #
            #self.add_mtx( mtx = link_mtx, dof_map = link_dof_map )

        return

    def get_dof_ix_array(self, dof):
        '''Get the access to the posisions containing values
        related to the dof
        '''
        return [mtx_array.get_dof_ix_array(dof)
                for mtx_array in self.get_sys_mtx_arrays()]

    def _zero_rc(self, dof_ix_arrays):
        for mtx_array, dof_ix_array in zip(self.get_sys_mtx_arrays(), dof_ix_arrays):
            mtx_array._zero_rc(dof_ix_array)

    def _add_col_to_vector(self, dof_ix_arrays, F, factor):
        for mtx_array, dof_ix_array in zip(self.get_sys_mtx_arrays(), dof_ix_arrays):
            mtx_array._add_col_to_vector(dof_ix_array, F, factor)

    def _get_col_subvector(self, dof_ix_arrays):
        idx_arr = np.array([], dtype=int)
        val_arr = np.array([], dtype=float)
        for mtx_array, dof_ix_array in zip(self.get_sys_mtx_arrays(), dof_ix_arrays):
            idx_seg, val_seg = mtx_array._get_col_subvector(dof_ix_array)
            idx_arr = np.append(idx_arr, idx_seg)
            val_arr = np.append(val_arr, val_seg)
        return idx_arr, val_arr

    def _get_diag_elem(self, dof_ix_arrays):
        K_dof_dof = 0
        for mtx_array, dof_ix_array in zip(self.get_sys_mtx_arrays(), dof_ix_arrays):
            K_dof_dof += mtx_array._get_diag_elem(dof_ix_array)
        return K_dof_dof

    def _add_diag_elem(self, dof_ix_arrays, K_aa):
        '''Set value of a diagonal element. Find the first occurrence
        of the element value and put it there
        '''
        for mtx_array, dof_ix_array in zip(self.get_sys_mtx_arrays(), dof_ix_arrays):
            el, r_ix = dof_ix_array
            if el.shape[0] > 0:
                mtx_array._add_diag_elem(dof_ix_array, K_aa)
                break
